"""@namespace IMP.pmi1.io.xltable
   Tools to plot a contact map overlaid with cross-links.
"""

from __future__ import print_function
from math import sqrt
from scipy.spatial.distance import cdist
import numpy as np
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.cm as cm
import matplotlib.pyplot as plt

from collections import defaultdict
import pickle
import IMP
import IMP.atom
import IMP.rmf
import RMF
import IMP.pmi1
import IMP.pmi1.io
import IMP.pmi1.io.crosslink
import IMP.pmi1.io.utilities
import IMP.pmi1.topology
import IMP.pmi1.analysis
import operator

@IMP.deprecated_object("2.8", "Use PMI2, IMP.pmi1.macros.AnalysisReplicaExchange,  instead")
class XLTable():
    """ class to read, analyze, and plot xlink data on contact maps
    Canonical way to read the data:
    1) load sequences and name them
    2) load coordinates for those sequences from PDB file
    3) add crosslinks
    4) create contact map
    5) plot
    """

    def __init__(self,contact_threshold):
        self.sequence_dict={}
        self.cross_link_db = None
        self.residue_pair_list = []          # list of special residue pairs to display
        self.distance_maps = []              # distance map for each copy of the complex
        self.contact_freqs = None
        self.num_pdbs = 0
        self.num_rmfs = 0
        self.index_dict = defaultdict(list)  # location in the dmap of each residue
        self.contact_threshold = contact_threshold
        # internal things
        self._first = True
        self.index_dict={}
        self.stored_dists={}
        self.model = IMP.Model()

    @property
    @IMP.deprecated_method("3.0", "Model should be accessed with `.model`.")
    def mdl(self):
        return self.model

    def _colormap_distance(self, dist, threshold=35, tolerance=0):
        if dist < threshold - tolerance:
            return "Green"
        elif dist >= threshold + tolerance:
            return "Orange"
        else:
            return "Red"

    def _colormap_satisfaction(self, sat, threshold=0.5, tolerance=0.1):
        if sat >=  threshold + tolerance:
            print(sat, "green")
            return "Green"
        elif sat < threshold + tolerance and sat >= threshold - tolerance :
            print(sat, "orange")
            return "Orange"
        else:
            print(sat, "orange")
            return "Orange"

    def _get_percentage_satisfaction(self,r1,c1,r2,c2,threshold=35):
        try:
            idx1=self.index_dict[c1][r1]
        except:
            return None
        try:
            idx2=self.index_dict[c2][r2]
        except:
            return None
        nsatisfied=0
        for dists in self.dist_maps:
            dist=dists[idx1,idx2]
            if dist<threshold: nsatisfied+=1
        return float(nsatisfied)/len(self.dist_maps)

    def _get_distance(self,r1,c1,r2,c2):
        if self.index_dict is not None:
            try:
                idx1=self.index_dict[c1][r1]
            except:
                return None
            try:
                idx2=self.index_dict[c2][r2]
            except:
                return None
            return self.av_dist_map[idx1,idx2]
        else:
            if (r1,c1,r2,c2) not in self.stored_dists.keys():
                sel=IMP.atom.Selection(self.prots,molecule=c1,residue_index=r1)
                selpart=sel.get_selected_particles()
                selpart_res_one=list(set(self.particles_resolution_one) & set(selpart))
                if len(selpart_res_one)>1: return None
                if len(selpart_res_one)==0: return None
                selpart_res_one_1=selpart_res_one[0]
                sel=IMP.atom.Selection(self.prots,molecule=c2,residue_index=r2)
                selpart=sel.get_selected_particles()
                selpart_res_one=list(set(self.particles_resolution_one) & set(selpart))
                if len(selpart_res_one)>1: return None
                if len(selpart_res_one)==0: return None
                selpart_res_one_2=selpart_res_one[0]
                d1=IMP.core.XYZ(selpart_res_one_1)
                d2=IMP.core.XYZ(selpart_res_one_2)
                dist=IMP.core.get_distance(d1,d2)
                self.stored_dists[(r1,c1,r2,c2)]=dist
            else:
                dist=self.stored_dists[(r1,c1,r2,c2)]
            return dist




    def _internal_load_maps(self,maps_fn):
        npzfile = np.load(maps_fn)
        cname_array=npzfile['cname_array']
        idx_array=npzfile['idx_array']
        index_dict={}
        for cname,idxs in zip(cname_array,idx_array):
            tmp=list(idxs)
            if -1 in tmp:
                index_dict[cname]=tmp[0:tmp.index(-1)]
            else:
                index_dict[cname]=tmp
        av_dist_map = npzfile['av_dist_map']
        contact_map = npzfile['contact_map']
        return index_dict,av_dist_map,contact_map

    def load_sequence_from_fasta_file(self,fasta_file,id_in_fasta_file,protein_name):
        """ read sequence. structures are displayed in the same order as sequences are read.
        fasta_file:        file to read
        id_in_fasta_file:  id of desired sequence
        protein_name:      identifier for this sequence (use same name when handling coordinates)
        can provide the fasta name (for retrieval) and the protein name (for storage) """

        record_dict = IMP.pmi1.topology.Sequences(fasta_file)
        if id_in_fasta_file is None:
            id_in_fasta_file = name
        if id_in_fasta_file not in record_dict:
            raise KeyError("id %s not found in fasta file" % id_in_fasta_file)
        length = len(record_dict[id_in_fasta_file])

        self.sequence_dict[protein_name] = str(record_dict[id_in_fasta_file])

    def load_pdb_coordinates(self,pdbfile,chain_to_name_map):
        """ read coordinates from a pdb file. also appends to distance maps
        @param pdbfile             file for reading coords
        @param chain_to_name_map   correspond chain ID with protein name (will ONLY read these chains)
                                   Key: PDB chain ID. Value: Protein name (set in sequence reading)
        \note This function returns an error if the sequence for each chain has NOT been read
        """
        mh = IMP.atom.read_pdb(pdbfile,self.model,IMP.atom.CAlphaPDBSelector())
        total_len = sum(len(self.sequence_dict[s]) for s in self.sequence_dict)
        coords = np.ones((total_len,3)) * 1e5 #default to coords "very far away"
        prev_stop = 0
        for cid in chain_to_name_map:
            cname = chain_to_name_map[cid]
            if cname not in self.sequence_dict:
                print("ERROR: chain",cname,'has not been read or has a naming mismatch')
                return
            if self._first:
                self.index_dict[cname]=range(prev_stop,prev_stop+len(self.sequence_dict[cname]))
            sel = IMP.atom.Selection(mh,chain_id=cid)
            for p in sel.get_selected_particles():
                rnum = IMP.atom.get_residue(IMP.atom.Atom(p)).get_index()
                coords[rnum+prev_stop-1,:] = list(IMP.core.XYZ(p).get_coordinates())
            prev_stop+=len(self.sequence_dict[cname])
        dists = cdist(coords, coords)
        binary_dists = np.where((dists <= self.contact_threshold) & (dists >= 1.0), 1.0, 0.0)
        if self._first:
            self.dist_maps= [dists]
            self.av_dist_map = dists
            self.contact_freqs = binary_dists
            self._first=False
        else:
            self.dist_maps.append(dists)
            self.av_dist_map += dists
            self.contact_freqs += binary_dists
        IMP.atom.destroy(mh)
        del mh
        self.num_pdbs+=1

    def load_rmf_coordinates(self,rmf_name,rmf_frame_index, chain_names, nomap=False):
        """ read coordinates from a rmf file. It needs IMP to run.
        rmf has been created using IMP.pmi1 conventions. It gets the
        highest resolution automatically. Also appends to distance maps
        @param rmf_name             file for reading coords
        @param rmf_frame_index      frame index from the rmf
        @param nomap                Default False, if True it will not calculate the contact map
        """
        (particles_resolution_one, prots)=self._get_rmf_structure(rmf_name,rmf_frame_index)

        total_len = sum(len(self.sequence_dict[s]) for s in self.sequence_dict)
        print(self.sequence_dict,total_len)

        coords = np.ones((total_len,3)) * 1e6 #default to coords "very far away"
        prev_stop=0
        sorted_particles=IMP.pmi1.tools.sort_by_residues(particles_resolution_one)

        self.prots=prots
        self.particles_resolution_one=particles_resolution_one

        if nomap:
            return

        for cname in chain_names:
            print(cname)
            if self._first:
                self.index_dict[cname]=range(prev_stop,prev_stop+len(self.sequence_dict[cname]))
            rindexes=range(1,len(self.sequence_dict[cname])+1)
            for rnum in rindexes:
                sel=IMP.atom.Selection(prots,molecule=cname,residue_index=rnum)
                selpart=sel.get_selected_particles()
                selpart_res_one=list(set(particles_resolution_one) & set(selpart))
                if len(selpart_res_one)>1: continue
                if len(selpart_res_one)==0: continue
                selpart_res_one=selpart_res_one[0]
                try:
                    coords[rnum+prev_stop-1,:]=IMP.core.XYZ(selpart_res_one).get_coordinates()
                except IndexError:
                    print("Error: exceed max size",prev_stop,total_len,cname,rnum)
                    exit()
            prev_stop+=len(self.sequence_dict[cname])
        dists = cdist(coords, coords)
        binary_dists = np.where((dists <= self.contact_threshold) & (dists >= 1.0), 1.0, 0.0)
        if self._first:
            self.dist_maps= [dists]
            self.av_dist_map = dists
            self.contact_freqs = binary_dists
            self._first=False
        else:
            self.dist_maps.append(dists)
            self.av_dist_map += dists
            self.contact_freqs += binary_dists
        self.num_rmfs+=1


    def _get_rmf_structure(self,rmf_name,rmf_frame_index):
        rh= RMF.open_rmf_file_read_only(rmf_name)
        prots=IMP.rmf.create_hierarchies(rh, self.model)
        IMP.rmf.load_frame(rh, rmf_frame_index)
        print("getting coordinates for frame %i rmf file %s" % (rmf_frame_index, rmf_name))
        del rh

        particle_dict=IMP.pmi1.analysis.get_particles_at_resolution_one(prots[0])

        protein_names=particle_dict.keys()
        particles_resolution_one=[]
        for k in particle_dict:
            particles_resolution_one+=(particle_dict[k])

        return particles_resolution_one, prots


    def save_maps(self,maps_fn):
        maxlen=max(len(self.index_dict[key]) for key in self.index_dict)
        cnames=[]
        idxs=[]
        for cname,idx in self.index_dict.items():
            cnames.append(cname)
            idxs.append(idx+[-1]*(maxlen-len(idx)))
        idx_array=np.array(idxs)
        cname_array=np.array(cnames)
        np.savez(maps_fn,
                 cname_array=cname_array,
                 idx_array=idx_array,
                 av_dist_map=self.av_dist_map,
                 contact_map=self.contact_freqs)

    def load_maps(self,maps_fn):
        self.index_dict,self.av_dist_map,self.contact_freqs=self._internal_load_maps(maps_fn)

    def load_crosslinks(self,CrossLinkDataBase,distance_field=None):
        """ read crosslinks from a CSV file.
        provide a CrossLinkDataBaseKeywordsConverter to explain the columns
        @distance_field is the optional keyword for the distance to be read form the file.
        This can skip the rmf reading to calculate the distance of cross-links if
        already provided in the csv file."""
        if type(CrossLinkDataBase) is not IMP.pmi1.io.crosslink.CrossLinkDataBase:
            raise TypeError("Crosslink database must be a IMP.pmi1.io.CrossLinkDataBase type")
        self.cross_link_db=CrossLinkDataBase
        if distance_field is not None:
            total_len = sum(len(self.sequence_dict[s]) for s in self.sequence_dict)
            zeros = np.zeros((total_len,3))
            self.av_dist_map = cdist(zeros,zeros)
            for xl in self.cross_link_db:
                c1=xl[self.cross_link_db.protein1_key]
                c2=xl[self.cross_link_db.protein2_key]
                r1=xl[self.cross_link_db.residue1_key]
                r2=xl[self.cross_link_db.residue2_key]
                try:
                    self.stored_dists[(r1,c1,r2,c2)]=float(xl[distance_field])
                    self.stored_dists[(r2,c2,r1,c1)]=float(xl[distance_field])
                except ValueError:
                    self.stored_dists[(r1,c1,r2,c2)]=10e6
                    self.stored_dists[(r2,c2,r1,c1)]=10e6
        else:
            for xl in self.cross_link_db:
                c1=xl[self.cross_link_db.protein1_key]
                c2=xl[self.cross_link_db.protein2_key]
                r1=xl[self.cross_link_db.residue1_key]
                r2=xl[self.cross_link_db.residue2_key]
                self.stored_dists[(r1,c1,r2,c2)]=self._get_distance(r1,c1,r2,c2)
                self.stored_dists[(r2,c2,r1,c1)]=self._get_distance(r2,c2,r1,c1)

    def set_residue_pairs_to_display(self,residue_type_pair):
        """ select the atom names of residue pairs to plot on the contact map
        list of residues types must be single letter code
        e.g. residue_type_pair=("K","K")
        """
        rtp=sorted(residue_type_pair)
        for prot1 in self.sequence_dict:
            seq1=self.sequence_dict[prot1]
            for nres1,res1 in enumerate(seq1):
                for prot2 in self.sequence_dict:
                    seq2=self.sequence_dict[prot2]
                    for nres2,res2 in enumerate(seq2):
                        if sorted((res1,res2))==rtp:
                            self.residue_pair_list.append((nres1+1,prot1,nres2+1,prot2))

    def setup_contact_map(self):
        """ loop through each distance map and get frequency of contacts
        """
        if self.num_pdbs!=0 and self.num_rmfs==0:
            self.av_dist_map = 1.0/self.num_pdbs * self.av_dist_map
            self.contact_freqs = 1.0/self.num_pdbs * self.contact_freqs
        if self.num_pdbs==0 and self.num_rmfs!=0:
            self.av_dist_map = 1.0/self.num_rmfs * self.av_dist_map
            self.contact_freqs = 1.0/self.num_rmfs * self.contact_freqs

    def setup_difference_map(self,maps_fn1,maps_fn2,thresh):
        idx1,av1,contact1=self._internal_load_maps(maps_fn1)
        idx2,av2,contact2=self._internal_load_maps(maps_fn2)
        if idx1!=idx2:
            print("UH OH: index dictionaries do not match!")
            exit()
        self.index_dict=idx1
        self.av_dist_map=av1 # should we store both somehow? only needed for XL

        def logic(c1,c2):
            if c1==0 and c2==0:             # white
                return 0
            elif c1>thresh and c2<thresh:   # red
                return 1
            elif c1<thresh and c2>thresh:   # blue
                return 2
            else:                           # green
                return 3
        f = np.vectorize(logic,otypes=[np.int])
        print('computing contact map')
        self.contact_freqs = f(contact1,contact2)
        print('done')



    def spring_layout(self,ax,plt,data, annotations, iterations = 100, k=1):

        import networkx as nx
        import numpy
        import scipy.spatial


        #import matplotlib.pyplot as plt

        """
        Author: G. Bouvier, Pasteur Institute, Paris
        Website: http://bloggb.fr/2015/10/19/spring_layout_for_annotating_plot_in_matplotlib.html
        - data: coordinates of your points [(x1,y1), (x2,y2), ..., (xn,yn)]
        - annotations: text for your annotation ['str_1', 'str_2', ..., 'str_n']
        - iterations: number of iterations for spring layout
        - k: optimal distance between nodes
        """
        G = nx.Graph()
        init_pos = {} # initial position
        x, y = [e[0] for e in data], [e[1] for e in data]
        xmin, xmax = min(x), max(x)
        ymin, ymax = min(y), max(y)

        m=IMP.Model()
        particles={}
        particles_fixed={}
        mvs = []

        rs = IMP.RestraintSet(m, 'distance')

        for i,xy in enumerate(data):
            p=IMP.Particle(m)
            G.add_node(i)
            init_pos[i] = xy
            IMP.core.XYZR.setup_particle(p)
            IMP.core.XYZ(p).set_coordinates((xy[0],xy[1],0))
            IMP.core.XYZR(p).set_radius(100)
            IMP.core.XYZ(p).set_coordinates_are_optimized(True)
            particles[i]=p
            Xfloatkey = IMP.core.XYZ.get_xyz_keys()[0]
            Yfloatkey = IMP.core.XYZ.get_xyz_keys()[1]
            mvs.append(IMP.core.NormalMover([p], [Xfloatkey], 5))
            mvs.append(IMP.core.NormalMover([p], [Yfloatkey], 5))
            p=IMP.Particle(m)
            IMP.core.XYZR.setup_particle(p)
            IMP.core.XYZ(p).set_coordinates((xy[0],xy[1],0))
            IMP.core.XYZR(p).set_radius(100)
            particles_fixed[i]=p
            ts1 = IMP.core.HarmonicUpperBound(200, 0.001)
            rs.add_restraint(
            IMP.core.DistanceRestraint(m, ts1,
                                       particles[i],
                                       particles_fixed[i]))

        ssps = IMP.core.SoftSpherePairScore(1.0)
        lsa = IMP.container.ListSingletonContainer(m)

        ps = particles.values()+particles_fixed.values()

        lsa.add(IMP.get_indexes(ps))

        rbcpf = IMP.core.RigidClosePairsFinder()
        cpc = IMP.container.ClosePairContainer(lsa, 0.0, rbcpf, 10.0)
        evr = IMP.container.PairsRestraint(ssps, cpc)
        rs.add_restraint(evr)

        '''
            #G.add_node(text)
            #G.add_edge(xy, text)
            #init_pos[text] = xy

        delTri = scipy.spatial.Delaunay(data)
        #plt.scatter([e[0] for e in data], [e[1] for e in data])
        # create a set for edges that are indexes of the points
        edges = set()
        # for each Delaunay triangle
        for n in xrange(delTri.nsimplex):
            # for each edge of the triangle
            # sort the vertices
            # (sorting avoids duplicated edges being added to the set)
            # and add to the edges set
            edge = sorted([delTri.vertices[n,0], delTri.vertices[n,1]])
            edges.add((edge[0], edge[1]))
            edge = sorted([delTri.vertices[n,0], delTri.vertices[n,2]])
            edges.add((edge[0], edge[1]))
            edge = sorted([delTri.vertices[n,1], delTri.vertices[n,2]])
            edges.add((edge[0], edge[1]))


        for e in edges:
            G.add_edge(e[0],e[1])

            d1=IMP.core.XYZ(particles[e[0]])
            d2=IMP.core.XYZ(particles[e[1]])
            dist=IMP.core.get_distance(d1,d2)
            ts1 = IMP.core.Harmonic(dist+150, 0.001)
            rs.add_restraint(
            IMP.core.DistanceRestraint(m, ts1,
                                       particles[e[0]],
                                       particles[e[1]]))
        '''

        smv = IMP.core.SerialMover(mvs)

        mc = IMP.core.MonteCarlo(m)
        mc.set_scoring_function(rs)
        mc.set_return_best(False)
        mc.set_kt(1.0)
        mc.add_mover(smv)


        '''
        pos = nx.spring_layout(G ,pos=init_pos)
        x_list=[]
        y_list=[]
        for i,xy in enumerate(data):
            xynew =  pos[i] * [xmax-xmin, ymax-ymin] + [xmin, ymin]
            plt.plot((xy[0],xynew[0]), (xy[1],xynew[1]), 'k-')
            x_list.append(xynew[0])
            y_list.append(xynew[1])
            #plt.annotate(name, xy, size=5,  xycoords='data', xytext=xytext, textcoords='data', bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),\
            #            arrowprops=dict(arrowstyle="-", connectionstyle="arc3", color='gray'))

        points=ax.scatter(x_list,y_list,alpha=1)
        #pos*=[numpy.ptp([e[0] for e in data]), numpy.ptp([e[1] for e in data])]
        #plt.show()
        return points
        '''

        for i in range(10):
            mc.optimize(1000*len(particles.keys())*2)
            print(rs.evaluate(False))
        x_list=[]
        y_list=[]
        for i,xy in enumerate(data):
            p=particles[i]
            coord=IMP.core.XYZ(p).get_coordinates()
            xynew =  (coord[0],coord[1])
            plt.plot((xy[0],xynew[0]), (xy[1],xynew[1]), 'k-')
            x_list.append(xynew[0])
            y_list.append(xynew[1])
        points=ax.scatter(x_list,y_list,alpha=1,facecolors='none', edgecolors='k')
        return points



    def show_mpld3(self,fig,ax,points,xl_list,xl_labels):
        import mpld3
        from mpld3 import plugins
        import pandas as pd

        # Define some CSS to control our custom labels
        css = """
        table
        {
          border-collapse: collapse;
        }
        th
        {
          color: #000000;
          background-color: #ffffff;
        }
        td
        {
          background-color: #cccccc;
        }
        table, th, td
        {
          font-family:Arial, Helvetica, sans-serif;
          border: 1px solid black;
          text-align: right;
          font-size: 10px;
        }
        """
        df = pd.DataFrame(index=xl_labels)

        sorted_keys=sorted(xl_list[0].keys())

        for k in sorted_keys:
            df[k] = np.array([xl[k] for xl in xl_list])

        labels = []
        for i in range(len(xl_labels)):
            label = df.ix[[i], :].T
            # .to_html() is unicode; so make leading 'u' go away with str()
            labels.append(str(label.to_html()))

        tooltip = plugins.PointHTMLTooltip(points, labels,
                                   voffset=10, hoffset=10, css=css)
        plugins.connect(fig, tooltip)
        mpld3.save_html(fig,"output.html")


    def get_residue_contacts(self,prot_listx,prot_listy):
        for px in prot_listx:
            for py in prot_listy:
                indexes_x = self.index_dict[px]
                minx = min(indexes_x)
                maxx = max(indexes_x)
                indexes_y = self.index_dict[py]
                miny = min(indexes_y)
                maxy = max(indexes_y)
                array = self.contact_freqs[minx:maxx,miny:maxy]
                (xresidues,yresidues)=np.where(array>0)
                for n,xr in enumerate(xresidues):
                    print(xr,yresidues[n],px,py,array[xr,yresidues[n]])


    def plot_table(self, prot_listx=None,
                   prot_listy=None,
                   no_dist_info=False,
                   confidence_info=False,
                   filter=None,
                   display_residue_pairs=False,
                   contactmap=False,
                   filename=None,
                   confidence_classes=None,
                   alphablend=0.1,
                   scale_symbol_size=1.0,
                   gap_between_components=0,
                   dictionary_of_gaps={},
                   colormap=cm.binary,
                   crosslink_threshold=None,
                   colornorm=None,
                   cbar_labels=None,
                   labels=False,
                   show_mpld3=False,
                   color_crosslinks_by_distance=True):
        """ plot the xlink table with optional contact map.
        prot_listx:             list of protein names on the x-axis
        prot_listy:             list of protein names on the y-axis
        no_dist_info:           plot only the cross-links as grey spots
        confidence_info:
        filter:                 list of tuples to filter on. each one contains:
                                    keyword in the database to be filtered on
                                    relationship ">","==","<"
                                    a value
                                example ("ID_Score",">",40)
        display_residue_pairs:  display all pairs defined in self.residue_pair_list
        contactmap:             display the contact map
        filename:               save to file (adds .pdf extension)
        confidence_classes:
        alphablend:
        scale_symbol_size:      rescale the symbol for the crosslink
        gap_between_components: the numbeber of residues to leave blannk between each component
        dictionary_of_gaps:     add extra space after the given protein. dictionary_of_gaps={prot_name:extra_gap}
        """
        # prepare figure


        fig = plt.figure(figsize=(100, 100))
        ax = fig.add_subplot(111)
        ax.set_xticks([])
        ax.set_yticks([])

        if cbar_labels is not None:
            if len(cbar_labels)!=4:
                print("to provide cbar labels, give 3 fields (first=first input file, last=last input) in oppose order of input contact maps")
                exit()
        # set the list of proteins on the x axis
        if prot_listx is None:
            prot_listx = self.sequence_dict.keys()
            prot_listx.sort()
        nresx = gap_between_components + \
            sum([len(self.sequence_dict[name])
                + gap_between_components for name in prot_listx]) + \
            sum([dictionary_of_gaps[prot] for prot in dictionary_of_gaps.keys()])

        # set the list of proteins on the y axis
        if prot_listy is None:
            prot_listy = self.sequence_dict.keys()
            prot_listy.sort()
        nresy = gap_between_components + \
            sum([len(self.sequence_dict[name])
                + gap_between_components for name in prot_listy]) + \
            sum([dictionary_of_gaps[prot] for prot in dictionary_of_gaps.keys()])

        # this is the residue offset for each protein
        resoffsetx = {}
        resendx = {}
        res = gap_between_components
        for prot in prot_listx:
            resoffsetx[prot] = res
            res += len(self.sequence_dict[prot])
            resendx[prot] = res
            res += gap_between_components
            if prot in dictionary_of_gaps.keys(): res+= dictionary_of_gaps[prot]

        resoffsety = {}
        resendy = {}
        res = gap_between_components
        for prot in prot_listy:
            resoffsety[prot] = res
            res += len(self.sequence_dict[prot])
            resendy[prot] = res
            res += gap_between_components
            if prot in dictionary_of_gaps.keys(): res+= dictionary_of_gaps[prot]

        resoffsetdiagonal = {}
        res = gap_between_components
        for prot in IMP.pmi1.io.utilities.OrderedSet(prot_listx + prot_listy):
            resoffsetdiagonal[prot] = res
            res += len(self.sequence_dict[prot])
            res += gap_between_components
            if prot in dictionary_of_gaps.keys(): res+= dictionary_of_gaps[prot]

        # plot protein boundaries
        xticks = []
        xlabels = []
        for n, prot in enumerate(prot_listx):
            res = resoffsetx[prot]
            end = resendx[prot]
            for proty in prot_listy:
                resy = resoffsety[proty]
                endy = resendy[proty]
                ax.plot([res, res], [resy, endy], linestyle='-',color='gray', lw=0.4)
                ax.plot([end, end], [resy, endy], linestyle='-',color='gray', lw=0.4)
            xticks.append((float(res) + float(end)) / 2)
            xlabels.append(prot)

        yticks = []
        ylabels = []
        for n, prot in enumerate(prot_listy):
            res = resoffsety[prot]
            end = resendy[prot]
            for protx in prot_listx:
                resx = resoffsetx[protx]
                endx = resendx[protx]
                ax.plot([resx, endx], [res, res], linestyle='-',color='gray', lw=0.4)
                ax.plot([resx, endx], [end, end], linestyle='-',color='gray', lw=0.4)
            yticks.append((float(res) + float(end)) / 2)
            ylabels.append(prot)

        # plot the contact map
        if contactmap:
            tmp_array = np.zeros((nresx, nresy))
            for px in prot_listx:
                for py in prot_listy:
                    resx = resoffsetx[px]
                    lengx = resendx[px] - 1
                    resy = resoffsety[py]
                    lengy = resendy[py] - 1
                    indexes_x = self.index_dict[px]
                    minx = min(indexes_x)
                    maxx = max(indexes_x)
                    indexes_y = self.index_dict[py]
                    miny = min(indexes_y)
                    maxy = max(indexes_y)
                    tmp_array[resx:lengx,resy:lengy] = self.contact_freqs[minx:maxx,miny:maxy]

            cax = ax.imshow(tmp_array,
                      cmap=colormap,
                      norm=colornorm,
                      origin='lower',
                      alpha=0.6,
                      interpolation='nearest')

        ax.set_xticks(xticks)
        ax.set_xticklabels(xlabels, rotation=90)
        ax.set_yticks(yticks)
        ax.set_yticklabels(ylabels)
        plt.setp(ax.get_xticklabels(), fontsize=6)
        plt.setp(ax.get_yticklabels(), fontsize=6)

        # set the crosslinks
        already_added_xls = []
        xl_coordinates_tuple_list = []
        xl_labels = []
        x_list=[]
        y_list=[]
        color_list=[]
        xl_list=[]

        markersize = 5 * scale_symbol_size
        if self.cross_link_db:
            for xl in self.cross_link_db:

                (c1,c2,r1,r2)=IMP.pmi1.io.crosslink._ProteinsResiduesArray(xl)
                label=xl[self.cross_link_db.unique_sub_id_key]
                if color_crosslinks_by_distance:

                    try:
                        mdist=self._get_distance(r1,c1,r2,c2)
                        if mdist is None: continue
                        color = self._colormap_distance(mdist,threshold=crosslink_threshold)
                    except KeyError:
                        color="gray"

                else:

                    try:
                        ps=self._get_percentage_satisfaction(r1,c1,r2,c2)
                        if ps is None: continue
                        color = self._colormap_satisfaction(ps,threshold=0.2,tolerance=0.1)
                    except KeyError:
                        color="gray"

                try:
                    pos1 = r1 + resoffsetx[c1]
                except:
                    continue
                try:
                    pos2 = r2 + resoffsety[c2]
                except:
                    continue


                # everything below is used for plotting the diagonal
                # when you have a rectangolar plots
                pos_for_diagonal1 = r1 + resoffsetdiagonal[c1]
                pos_for_diagonal2 = r2 + resoffsetdiagonal[c2]
                if confidence_info:
                    if confidence == '0.01':
                        markersize = 14 * scale_symbol_size
                    elif confidence == '0.05':
                        markersize = 9 * scale_symbol_size
                    elif confidence == '0.1':
                        markersize = 6 * scale_symbol_size
                    else:
                        markersize = 15 * scale_symbol_size
                else:
                    markersize = 5 * scale_symbol_size
                '''
                ax.plot([pos1],
                        [pos2],
                        'o',
                        c=color,
                        alpha=alphablend,
                        markersize=markersize)

                ax.plot([pos2],
                        [pos1],
                        'o',
                        c=color,
                        alpha=alphablend,
                        markersize=markersize)
                '''
                x_list.append(pos1)
                x_list.append(pos2)
                y_list.append(pos2)
                y_list.append(pos1)
                color_list.append(color)
                color_list.append(color)
                xl["Distance"]=mdist
                xl_list.append(xl)
                xl_list.append(xl)

                xl_labels.append(label)
                xl_coordinates_tuple_list.append((float(pos1),float(pos2)))
                xl_labels.append(label+"*")
                xl_coordinates_tuple_list.append((float(pos2),float(pos1)))

            points=ax.scatter(x_list,y_list,s=markersize,c=color_list,alpha=alphablend)

        # plot requested residue pairs
        if display_residue_pairs:
            for rp in self.residue_pair_list:
                r1=rp[0]
                c1=rp[1]
                r2=rp[2]
                c2=rp[3]

                try:
                    dist=self._get_distance(r1,c1,r2,c2)
                except:
                    continue

                if dist<=40.0:
                    print(rp)
                    try:
                        pos1 = r1 + resoffsetx[c1]
                    except:
                        continue
                    try:
                        pos2 = r2 + resoffsety[c2]
                    except:
                        continue

                    ax.plot([pos1],
                            [pos2],
                            '+',
                            c="blue",
                            alpha=0.1,
                            markersize=markersize)

        # display and write to file
        fig.set_size_inches(0.002 * nresx, 0.002 * nresy)
        [i.set_linewidth(2.0) for i in ax.spines.values()]
        if cbar_labels is not None:
            cbar = fig.colorbar(cax, ticks=[0.5,1.5,2.5,3.5])
            cbar.ax.set_yticklabels(cbar_labels)# vertically oriented colorbar

        # set the labels using spring layout
        if labels:
            points=self.spring_layout(ax,plt,xl_coordinates_tuple_list, xl_labels)

        if filename:
            #plt.show()
            plt.savefig(filename, dpi=300,transparent="False")


        if show_mpld3:
            self.show_mpld3(fig,ax,points,xl_list,xl_labels)



## \example core/restrain_diameter.py
# An example restraining the diameter of a set of points. That is, the
# restraint penalizes conformations where there are two points more than a
# certain distance from one another.

import IMP
import IMP.core
import IMP.container
import sys

IMP.setup_from_argv(sys.argv, "restrain diameter")

# This example restraint the diameter of a set of particles to be smaller
# than 10

diameter = 10
m = IMP.Model()
lc = IMP.container.ListSingletonContainer(
               m, IMP.core.create_xyzr_particles(m, 50, 1.0))
h = IMP.core.HarmonicUpperBound(0, 1)
r = IMP.core.DiameterRestraint(h, lc, diameter)
sf = IMP.core.RestraintsScoringFunction([r])

# Set up optimizer
o = IMP.core.ConjugateGradients(m)
o.set_scoring_function(sf)

max = 0
for pi0 in lc.get_contents():
    for pi1 in lc.get_contents():
        d = IMP.core.get_distance(IMP.core.XYZ(m, pi0),
                                  IMP.core.XYZ(m, pi1))
        if d > max:
            max = d
print("The maximum distance is " + str(max))

IMP.set_log_level(IMP.SILENT)
o.optimize(100)

max = 0
for pi0 in lc.get_contents():
    for pi1 in lc.get_contents():
        d = IMP.core.get_distance(IMP.core.XYZ(m, pi0),
                                  IMP.core.XYZ(m, pi1))
        if d > max:
            max = d
print("Afterwards, the maximum distance is " + str(max))



## \example atom/cg_pdb.py
# This example shows how to properly write out a pdb of the structure
# results when using coarse grained rigid bodies.

from __future__ import print_function
import IMP.atom
import sys

IMP.setup_from_argv(sys.argv, "CG pdb")

m = IMP.Model()
full = IMP.atom.read_pdb(IMP.atom.get_example_path("example_protein.pdb"), m)
chain = IMP.atom.get_by_type(full, IMP.atom.CHAIN_TYPE)[0]
print(chain)
# for some reason the python wrapper won't make the implicit conversion to
# Chain
simplified = IMP.atom.create_simplified_along_backbone(
    IMP.atom.Chain(chain), 3)
IMP.atom.destroy(full)
rb = IMP.atom.create_rigid_body(simplified)
original_transform = rb.get_reference_frame().get_transformation_to()


# fake optimization, just move the rigid body
tr = IMP.algebra.get_random_rotation_3d()
rb.set_reference_frame(IMP.algebra.ReferenceFrame3D(
    IMP.algebra.Transformation3D(tr, IMP.algebra.Vector3D(0, 0, 0))))

# extract the difference
diff = rb.get_reference_frame().get_transformation_to() / original_transform
reload = IMP.atom.read_pdb(IMP.atom.get_example_path("example_protein.pdb"), m)
IMP.atom.transform(reload, diff)
name = IMP.create_temporary_file("out", ".pdb")
IMP.atom.write_pdb(reload, name)



## \example domino/multiscale.py
# We are interested in applying domino to problems systematically in a
# multiscale manner. This script experiments with those approaches.

from __future__ import print_function
import IMP.domino
import IMP.core
import sys

IMP.setup_from_argv(sys.argv, "multiscale")

m = IMP.Model()
m.set_log_level(IMP.SILENT)
ds = [IMP.core.XYZR.setup_particle(IMP.Particle(m))
      for i in range(0, 3)]
for i, d in enumerate(ds):
    d.set_radius(1)
    IMP.display.Colored.setup_particle(d, IMP.display.get_display_color(i))

k = 1
h = IMP.core.Harmonic(0, k)
r0 = IMP.core.SingletonRestraint(m,
    IMP.core.DistanceToSingletonScore(h, IMP.algebra.Vector3D(0, 0, 0)),
    ds[0].get_particle_index(), "0 at origin")

r1 = IMP.core.SingletonRestraint(m, IMP.core.AttributeSingletonScore(h,
                                                                  IMP.core.XYZ.get_xyz_keys(
                                                                  )[0]),
                                 ds[1].get_particle_index(), "1 on axis")

rs = [r0, r1]
for pr in [(0, 1), (1, 2), (0, 2)]:
    r = IMP.core.PairRestraint(m,
                 IMP.core.HarmonicSphereDistancePairScore(0, k),
                 (ds[pr[0]].get_particle_index(),
                  ds[pr[1]].get_particle_index()),
                 "R for " + str(pr))
    rs.append(r)


bb = IMP.algebra.BoundingBox2D(IMP.algebra.Vector2D(0, 0),
                               IMP.algebra.Vector2D(4, 4))

covers = []
for i in range(0, 6):
    cur = IMP.algebra.get_grid_interior_cover_by_spacing(bb, 4.0 / 2 ** i)
    print(cur)
    covers.append([IMP.algebra.Vector3D(x[0], x[1], 0) for x in cur])


def setup(cover, scale):
    pst = IMP.domino.ParticleStatesTable()
    st = IMP.domino.XYZStates(cover)
    for p in ds:
        pst.set_particle_states(p, st)
    for r in rs:
        r.set_maximum_score(.5 * scale ** 2)
    lf = IMP.domino.ListSubsetFilterTable(pst)
    rc = IMP.domino.RestraintCache(pst)
    rc.add_restraints(rs)
    fs = [IMP.domino.RestraintScoreSubsetFilterTable(rc),
          lf]
    sampler = IMP.domino.DominoSampler(m, pst)
    sampler.set_restraints(rs)
    sampler.set_subset_filter_tables(fs)
    sampler.set_log_level(IMP.SILENT)
    return (sampler, lf, pst)

(sampler, lf, pst) = setup(covers[0], 4.0)

subset = IMP.domino.Subset(ds)
ac = sampler.get_sample_assignments(subset)

print(ac)


def get_mapping(cover0, cover1):
    nn = IMP.algebra.NearestNeighbor3D(cover0)
    ret = [[] for c in cover0]
    for i, p in enumerate(cover1):
        nns = nn.get_nearest_neighbor(p)
        ret[nns].append(i)
    return ret

mw = IMP.display.PymolWriter("mapping.pym")


def display_mapping(index, cover0, cover1, mapping):
    mw.set_frame(index)
    for i, c in enumerate(mapping):
        for p in c:
            g = IMP.display.PointGeometry(cover1[p])
            g.set_color(IMP.display.get_display_color(i))
            g.set_name("fine")
            mw.add_geometry(g)
    for i, c in enumerate(cover0):
        g = IMP.display.PointGeometry(c)
        g.set_color(IMP.display.get_display_color(i))
        g.set_name("coarse")
        mw.add_geometry(g)


for curi in range(1, len(covers)):
    scale = 4.0 / 2 ** curi
    print(scale)
    mapping = get_mapping(covers[curi - 1], covers[curi])
    print(mapping)
    display_mapping(curi - 1, covers[curi - 1], covers[curi], mapping)
    (sampler, lf, pst) = setup(covers[curi], scale)
    lac = ac
    cac = []
    for a in lac:
        for i, p in enumerate(subset):
            s = a[i]
            allowed = mapping[s]
            lf.set_allowed_states(p, allowed)
        ccac = sampler.get_sample_assignments(subset)
        print(a, ccac)
        cac = cac + ccac
    ac = list(set(cac))
    print("for scale", scale, "got", ac)
    sw = IMP.display.PymolWriter("solutions." + str(curi) + ".pym")
    for i, a in enumerate(ac):
        IMP.domino.load_particle_states(subset, a, pst)
        sw.set_frame(i)
        for p in ds:
            g = IMP.core.XYZRGeometry(p)
            sw.add_geometry(g)
        for c in covers[curi]:
            g = IMP.display.PointGeometry(c)
            g.set_color(IMP.display.Color(1, 1, 1))
            g.set_name("grid")
            sw.add_geometry(g)



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

# This wrapper is part of IMP,
# Copyright 2007-2021 IMP Inventors. All rights reserved.

from __future__ import print_function, division, absolute_import
from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

import _IMP_rmf

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


import weakref

class IMP_RMF_SwigPyIterator(object):
    r"""Proxy of C++ swig::IMP_RMF_SwigPyIterator class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _IMP_rmf.delete_IMP_RMF_SwigPyIterator

    def value(self):
        r"""value(IMP_RMF_SwigPyIterator self) -> PyObject *"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_value(self)

    def incr(self, n=1):
        r"""incr(IMP_RMF_SwigPyIterator self, size_t n=1) -> IMP_RMF_SwigPyIterator"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        r"""decr(IMP_RMF_SwigPyIterator self, size_t n=1) -> IMP_RMF_SwigPyIterator"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_decr(self, n)

    def distance(self, x):
        r"""distance(IMP_RMF_SwigPyIterator self, IMP_RMF_SwigPyIterator x) -> ptrdiff_t"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_distance(self, x)

    def equal(self, x):
        r"""equal(IMP_RMF_SwigPyIterator self, IMP_RMF_SwigPyIterator x) -> bool"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_equal(self, x)

    def copy(self):
        r"""copy(IMP_RMF_SwigPyIterator self) -> IMP_RMF_SwigPyIterator"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_copy(self)

    def next(self):
        r"""next(IMP_RMF_SwigPyIterator self) -> PyObject *"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_next(self)

    def __next__(self):
        r"""__next__(IMP_RMF_SwigPyIterator self) -> PyObject *"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator___next__(self)

    def previous(self):
        r"""previous(IMP_RMF_SwigPyIterator self) -> PyObject *"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_previous(self)

    def advance(self, n):
        r"""advance(IMP_RMF_SwigPyIterator self, ptrdiff_t n) -> IMP_RMF_SwigPyIterator"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        r"""__eq__(IMP_RMF_SwigPyIterator self, IMP_RMF_SwigPyIterator x) -> bool"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        r"""__ne__(IMP_RMF_SwigPyIterator self, IMP_RMF_SwigPyIterator x) -> bool"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        r"""__iadd__(IMP_RMF_SwigPyIterator self, ptrdiff_t n) -> IMP_RMF_SwigPyIterator"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        r"""__isub__(IMP_RMF_SwigPyIterator self, ptrdiff_t n) -> IMP_RMF_SwigPyIterator"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        r"""__add__(IMP_RMF_SwigPyIterator self, ptrdiff_t n) -> IMP_RMF_SwigPyIterator"""
        return _IMP_rmf.IMP_RMF_SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        r"""
        __sub__(IMP_RMF_SwigPyIterator self, ptrdiff_t n) -> IMP_RMF_SwigPyIterator
        __sub__(IMP_RMF_SwigPyIterator self, IMP_RMF_SwigPyIterator x) -> ptrdiff_t
        """
        return _IMP_rmf.IMP_RMF_SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self

# Register IMP_RMF_SwigPyIterator in _IMP_rmf:
_IMP_rmf.IMP_RMF_SwigPyIterator_swigregister(IMP_RMF_SwigPyIterator)


_value_types=[]
_object_types=[]
_raii_types=[]
_plural_types=[]

IMP_DEBUG = _IMP_rmf.IMP_DEBUG

IMP_RELEASE = _IMP_rmf.IMP_RELEASE

IMP_SILENT = _IMP_rmf.IMP_SILENT

IMP_PROGRESS = _IMP_rmf.IMP_PROGRESS

IMP_TERSE = _IMP_rmf.IMP_TERSE

IMP_VERBOSE = _IMP_rmf.IMP_VERBOSE

IMP_MEMORY = _IMP_rmf.IMP_MEMORY

IMP_NONE = _IMP_rmf.IMP_NONE

IMP_USAGE = _IMP_rmf.IMP_USAGE

IMP_INTERNAL = _IMP_rmf.IMP_INTERNAL

IMP_KERNEL_HAS_LOG4CXX = _IMP_rmf.IMP_KERNEL_HAS_LOG4CXX

IMP_COMPILER_HAS_AUTO = _IMP_rmf.IMP_COMPILER_HAS_AUTO

IMP_COMPILER_HAS_DEBUG_VECTOR = _IMP_rmf.IMP_COMPILER_HAS_DEBUG_VECTOR

IMP_COMPILER_HAS_UNIQUE_PTR = _IMP_rmf.IMP_COMPILER_HAS_UNIQUE_PTR

IMP_KERNEL_HAS_BOOST_RANDOM = _IMP_rmf.IMP_KERNEL_HAS_BOOST_RANDOM

IMP_KERNEL_HAS_NUMPY = _IMP_rmf.IMP_KERNEL_HAS_NUMPY

IMP_KERNEL_HAS_GPERFTOOLS = _IMP_rmf.IMP_KERNEL_HAS_GPERFTOOLS

IMP_KERNEL_HAS_TCMALLOC_HEAPCHECKER = _IMP_rmf.IMP_KERNEL_HAS_TCMALLOC_HEAPCHECKER

IMP_KERNEL_HAS_TCMALLOC_HEAPPROFILER = _IMP_rmf.IMP_KERNEL_HAS_TCMALLOC_HEAPPROFILER

IMPKERNEL_SHOW_WARNINGS = _IMP_rmf.IMPKERNEL_SHOW_WARNINGS


import sys
class _DirectorObjects(object):
    """@internal Simple class to keep references to director objects
       to prevent premature deletion."""
    def __init__(self):
        self._objects = []
    def register(self, obj):
        """Take a reference to a director object; will only work for
           refcounted C++ classes"""
        if hasattr(obj, 'get_ref_count'):
            self._objects.append(obj)
    def cleanup(self):
        """Only drop our reference and allow cleanup by Python if no other
           Python references exist (we hold 3 references: one in self._objects,
           one in x, and one in the argument list for getrefcount) *and* no
           other C++ references exist (the Python object always holds one)"""
        objs = [x for x in self._objects if sys.getrefcount(x) > 3 \
                                         or x.get_ref_count() > 1]
# Do in two steps so the references are kept until the end of the
# function (deleting references may trigger a fresh call to this method)
        self._objects = objs
    def get_object_count(self):
        """Get number of director objects (useful for testing only)"""
        return len(self._objects)
_director_objects = _DirectorObjects()

class _ostream(object):
    r"""Proxy of C++ std::ostream class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr

    def write(self, osa_buf):
        r"""write(_ostream self, char const * osa_buf)"""
        return _IMP_rmf._ostream_write(self, osa_buf)

# Register _ostream in _IMP_rmf:
_IMP_rmf._ostream_swigregister(_ostream)

IMP_COMPILER_HAS_OVERRIDE = _IMP_rmf.IMP_COMPILER_HAS_OVERRIDE

IMP_COMPILER_HAS_FINAL = _IMP_rmf.IMP_COMPILER_HAS_FINAL

IMP_HAS_NOEXCEPT = _IMP_rmf.IMP_HAS_NOEXCEPT

IMP_C_OPEN_BINARY = _IMP_rmf.IMP_C_OPEN_BINARY

import IMP
IMP_CGAL_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_CGAL_HAS_BOOST_FILESYSTEM

IMP_CGAL_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_CGAL_HAS_BOOST_PROGRAMOPTIONS

IMP_CGAL_HAS_BOOST_RANDOM = _IMP_rmf.IMP_CGAL_HAS_BOOST_RANDOM

IMP_CGAL_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_CGAL_HAS_BOOST_SYSTEM

IMP_CGAL_HAS_NUMPY = _IMP_rmf.IMP_CGAL_HAS_NUMPY

IMPCGAL_SHOW_WARNINGS = _IMP_rmf.IMPCGAL_SHOW_WARNINGS

import IMP.cgal
IMP_ALGEBRA_HAS_IMP_CGAL = _IMP_rmf.IMP_ALGEBRA_HAS_IMP_CGAL

IMP_ALGEBRA_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_ALGEBRA_HAS_BOOST_FILESYSTEM

IMP_ALGEBRA_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_ALGEBRA_HAS_BOOST_PROGRAMOPTIONS

IMP_ALGEBRA_HAS_BOOST_RANDOM = _IMP_rmf.IMP_ALGEBRA_HAS_BOOST_RANDOM

IMP_ALGEBRA_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_ALGEBRA_HAS_BOOST_SYSTEM

IMP_ALGEBRA_HAS_CGAL = _IMP_rmf.IMP_ALGEBRA_HAS_CGAL

IMP_ALGEBRA_HAS_NUMPY = _IMP_rmf.IMP_ALGEBRA_HAS_NUMPY

IMP_ALGEBRA_HAS_ANN = _IMP_rmf.IMP_ALGEBRA_HAS_ANN

IMPALGEBRA_SHOW_WARNINGS = _IMP_rmf.IMPALGEBRA_SHOW_WARNINGS

import IMP.algebra
IMP_DISPLAY_HAS_IMP_CGAL = _IMP_rmf.IMP_DISPLAY_HAS_IMP_CGAL

IMP_DISPLAY_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_DISPLAY_HAS_BOOST_FILESYSTEM

IMP_DISPLAY_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_DISPLAY_HAS_BOOST_PROGRAMOPTIONS

IMP_DISPLAY_HAS_BOOST_RANDOM = _IMP_rmf.IMP_DISPLAY_HAS_BOOST_RANDOM

IMP_DISPLAY_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_DISPLAY_HAS_BOOST_SYSTEM

IMP_DISPLAY_HAS_CGAL = _IMP_rmf.IMP_DISPLAY_HAS_CGAL

IMP_DISPLAY_HAS_NUMPY = _IMP_rmf.IMP_DISPLAY_HAS_NUMPY

IMPDISPLAY_SHOW_WARNINGS = _IMP_rmf.IMPDISPLAY_SHOW_WARNINGS

import IMP.display
IMP_SCORE_FUNCTOR_HAS_IMP_CGAL = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_IMP_CGAL

IMP_SCORE_FUNCTOR_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_BOOST_FILESYSTEM

IMP_SCORE_FUNCTOR_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_BOOST_PROGRAMOPTIONS

IMP_SCORE_FUNCTOR_HAS_BOOST_RANDOM = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_BOOST_RANDOM

IMP_SCORE_FUNCTOR_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_BOOST_SYSTEM

IMP_SCORE_FUNCTOR_HAS_CGAL = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_CGAL

IMP_SCORE_FUNCTOR_HAS_HDF5 = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_HDF5

IMP_SCORE_FUNCTOR_HAS_NUMPY = _IMP_rmf.IMP_SCORE_FUNCTOR_HAS_NUMPY

IMPSCOREFUNCTOR_SHOW_WARNINGS = _IMP_rmf.IMPSCOREFUNCTOR_SHOW_WARNINGS

import IMP.score_functor
IMP_STATISTICS_HAS_IMP_CGAL = _IMP_rmf.IMP_STATISTICS_HAS_IMP_CGAL

IMP_STATISTICS_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_STATISTICS_HAS_BOOST_FILESYSTEM

IMP_STATISTICS_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_STATISTICS_HAS_BOOST_PROGRAMOPTIONS

IMP_STATISTICS_HAS_BOOST_RANDOM = _IMP_rmf.IMP_STATISTICS_HAS_BOOST_RANDOM

IMP_STATISTICS_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_STATISTICS_HAS_BOOST_SYSTEM

IMP_STATISTICS_HAS_CGAL = _IMP_rmf.IMP_STATISTICS_HAS_CGAL

IMP_STATISTICS_HAS_NUMPY = _IMP_rmf.IMP_STATISTICS_HAS_NUMPY

IMPSTATISTICS_SHOW_WARNINGS = _IMP_rmf.IMPSTATISTICS_SHOW_WARNINGS

import IMP.statistics
IMP_CORE_HAS_IMP_CGAL = _IMP_rmf.IMP_CORE_HAS_IMP_CGAL

IMP_CORE_HAS_IMP_KERNEL = _IMP_rmf.IMP_CORE_HAS_IMP_KERNEL

IMP_CORE_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_CORE_HAS_BOOST_FILESYSTEM

IMP_CORE_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_CORE_HAS_BOOST_PROGRAMOPTIONS

IMP_CORE_HAS_BOOST_RANDOM = _IMP_rmf.IMP_CORE_HAS_BOOST_RANDOM

IMP_CORE_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_CORE_HAS_BOOST_SYSTEM

IMP_CORE_HAS_CGAL = _IMP_rmf.IMP_CORE_HAS_CGAL

IMP_CORE_HAS_HDF5 = _IMP_rmf.IMP_CORE_HAS_HDF5

IMP_CORE_HAS_NUMPY = _IMP_rmf.IMP_CORE_HAS_NUMPY

IMPCORE_SHOW_WARNINGS = _IMP_rmf.IMPCORE_SHOW_WARNINGS

import IMP.core
IMP_CONTAINER_HAS_IMP_ALGEBRA = _IMP_rmf.IMP_CONTAINER_HAS_IMP_ALGEBRA

IMP_CONTAINER_HAS_IMP_CGAL = _IMP_rmf.IMP_CONTAINER_HAS_IMP_CGAL

IMP_CONTAINER_HAS_IMP_DISPLAY = _IMP_rmf.IMP_CONTAINER_HAS_IMP_DISPLAY

IMP_CONTAINER_HAS_IMP_KERNEL = _IMP_rmf.IMP_CONTAINER_HAS_IMP_KERNEL

IMP_CONTAINER_HAS_IMP_SCORE_FUNCTOR = _IMP_rmf.IMP_CONTAINER_HAS_IMP_SCORE_FUNCTOR

IMP_CONTAINER_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_CONTAINER_HAS_BOOST_FILESYSTEM

IMP_CONTAINER_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_CONTAINER_HAS_BOOST_PROGRAMOPTIONS

IMP_CONTAINER_HAS_BOOST_RANDOM = _IMP_rmf.IMP_CONTAINER_HAS_BOOST_RANDOM

IMP_CONTAINER_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_CONTAINER_HAS_BOOST_SYSTEM

IMP_CONTAINER_HAS_CGAL = _IMP_rmf.IMP_CONTAINER_HAS_CGAL

IMP_CONTAINER_HAS_HDF5 = _IMP_rmf.IMP_CONTAINER_HAS_HDF5

IMP_CONTAINER_HAS_NUMPY = _IMP_rmf.IMP_CONTAINER_HAS_NUMPY

IMP_CONTAINER_HAS_PYTHON_IHM = _IMP_rmf.IMP_CONTAINER_HAS_PYTHON_IHM

IMP_CONTAINER_HAS_GOOGLE_DENSE_HASH_MAP = _IMP_rmf.IMP_CONTAINER_HAS_GOOGLE_DENSE_HASH_MAP

IMP_CONTAINER_HAS_ROBIN_MAP = _IMP_rmf.IMP_CONTAINER_HAS_ROBIN_MAP

IMPCONTAINER_SHOW_WARNINGS = _IMP_rmf.IMPCONTAINER_SHOW_WARNINGS

import IMP.container
IMP_ATOM_HAS_IMP_CGAL = _IMP_rmf.IMP_ATOM_HAS_IMP_CGAL

IMP_ATOM_HAS_IMP_KERNEL = _IMP_rmf.IMP_ATOM_HAS_IMP_KERNEL

IMP_ATOM_HAS_IMP_SCORE_FUNCTOR = _IMP_rmf.IMP_ATOM_HAS_IMP_SCORE_FUNCTOR

IMP_ATOM_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_ATOM_HAS_BOOST_PROGRAMOPTIONS

IMP_ATOM_HAS_BOOST_RANDOM = _IMP_rmf.IMP_ATOM_HAS_BOOST_RANDOM

IMP_ATOM_HAS_BOOST_REGEX = _IMP_rmf.IMP_ATOM_HAS_BOOST_REGEX

IMP_ATOM_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_ATOM_HAS_BOOST_SYSTEM

IMP_ATOM_HAS_CGAL = _IMP_rmf.IMP_ATOM_HAS_CGAL

IMP_ATOM_HAS_HDF5 = _IMP_rmf.IMP_ATOM_HAS_HDF5

IMP_ATOM_HAS_NUMPY = _IMP_rmf.IMP_ATOM_HAS_NUMPY

IMP_ATOM_HAS_PYTHON_IHM = _IMP_rmf.IMP_ATOM_HAS_PYTHON_IHM

IMPATOM_SHOW_WARNINGS = _IMP_rmf.IMPATOM_SHOW_WARNINGS

IMP_ATOM_TYPE_INDEX = _IMP_rmf.IMP_ATOM_TYPE_INDEX

IMP_RESIDUE_TYPE_INDEX = _IMP_rmf.IMP_RESIDUE_TYPE_INDEX

IMP_HIERARCHY_TYPE_INDEX = _IMP_rmf.IMP_HIERARCHY_TYPE_INDEX

IMP_CHAIN_TYPE_INDEX = _IMP_rmf.IMP_CHAIN_TYPE_INDEX

import IMP.atom
IMP_EM_HAS_IMP_CGAL = _IMP_rmf.IMP_EM_HAS_IMP_CGAL

IMP_EM_HAS_IMP_CONTAINER = _IMP_rmf.IMP_EM_HAS_IMP_CONTAINER

IMP_EM_HAS_IMP_DISPLAY = _IMP_rmf.IMP_EM_HAS_IMP_DISPLAY

IMP_EM_HAS_IMP_SCORE_FUNCTOR = _IMP_rmf.IMP_EM_HAS_IMP_SCORE_FUNCTOR

IMP_EM_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_EM_HAS_BOOST_FILESYSTEM

IMP_EM_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_EM_HAS_BOOST_PROGRAMOPTIONS

IMP_EM_HAS_BOOST_RANDOM = _IMP_rmf.IMP_EM_HAS_BOOST_RANDOM

IMP_EM_HAS_BOOST_REGEX = _IMP_rmf.IMP_EM_HAS_BOOST_REGEX

IMP_EM_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_EM_HAS_BOOST_SYSTEM

IMP_EM_HAS_CGAL = _IMP_rmf.IMP_EM_HAS_CGAL

IMP_EM_HAS_HDF5 = _IMP_rmf.IMP_EM_HAS_HDF5

IMP_EM_HAS_NUMPY = _IMP_rmf.IMP_EM_HAS_NUMPY

IMP_EM_HAS_PYTHON_IHM = _IMP_rmf.IMP_EM_HAS_PYTHON_IHM

IMPEM_SHOW_WARNINGS = _IMP_rmf.IMPEM_SHOW_WARNINGS

import IMP.em
IMP_SAXS_HAS_IMP_CGAL = _IMP_rmf.IMP_SAXS_HAS_IMP_CGAL

IMP_SAXS_HAS_IMP_CONTAINER = _IMP_rmf.IMP_SAXS_HAS_IMP_CONTAINER

IMP_SAXS_HAS_IMP_DISPLAY = _IMP_rmf.IMP_SAXS_HAS_IMP_DISPLAY

IMP_SAXS_HAS_IMP_KERNEL = _IMP_rmf.IMP_SAXS_HAS_IMP_KERNEL

IMP_SAXS_HAS_IMP_SCORE_FUNCTOR = _IMP_rmf.IMP_SAXS_HAS_IMP_SCORE_FUNCTOR

IMP_SAXS_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_SAXS_HAS_BOOST_FILESYSTEM

IMP_SAXS_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_SAXS_HAS_BOOST_PROGRAMOPTIONS

IMP_SAXS_HAS_BOOST_RANDOM = _IMP_rmf.IMP_SAXS_HAS_BOOST_RANDOM

IMP_SAXS_HAS_BOOST_REGEX = _IMP_rmf.IMP_SAXS_HAS_BOOST_REGEX

IMP_SAXS_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_SAXS_HAS_BOOST_SYSTEM

IMP_SAXS_HAS_CGAL = _IMP_rmf.IMP_SAXS_HAS_CGAL

IMP_SAXS_HAS_HDF5 = _IMP_rmf.IMP_SAXS_HAS_HDF5

IMP_SAXS_HAS_NUMPY = _IMP_rmf.IMP_SAXS_HAS_NUMPY

IMP_SAXS_HAS_PYTHON_IHM = _IMP_rmf.IMP_SAXS_HAS_PYTHON_IHM

IMPSAXS_SHOW_WARNINGS = _IMP_rmf.IMPSAXS_SHOW_WARNINGS

import IMP.saxs
IMP_ISD_HAS_IMP_CGAL = _IMP_rmf.IMP_ISD_HAS_IMP_CGAL

IMP_ISD_HAS_IMP_DISPLAY = _IMP_rmf.IMP_ISD_HAS_IMP_DISPLAY

IMP_ISD_HAS_IMP_SCORE_FUNCTOR = _IMP_rmf.IMP_ISD_HAS_IMP_SCORE_FUNCTOR

IMP_ISD_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_ISD_HAS_BOOST_FILESYSTEM

IMP_ISD_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_ISD_HAS_BOOST_PROGRAMOPTIONS

IMP_ISD_HAS_BOOST_RANDOM = _IMP_rmf.IMP_ISD_HAS_BOOST_RANDOM

IMP_ISD_HAS_BOOST_REGEX = _IMP_rmf.IMP_ISD_HAS_BOOST_REGEX

IMP_ISD_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_ISD_HAS_BOOST_SYSTEM

IMP_ISD_HAS_CGAL = _IMP_rmf.IMP_ISD_HAS_CGAL

IMP_ISD_HAS_HDF5 = _IMP_rmf.IMP_ISD_HAS_HDF5

IMP_ISD_HAS_NUMPY = _IMP_rmf.IMP_ISD_HAS_NUMPY

IMP_ISD_HAS_PYTHON_IHM = _IMP_rmf.IMP_ISD_HAS_PYTHON_IHM

IMPISD_SHOW_WARNINGS = _IMP_rmf.IMPISD_SHOW_WARNINGS

import IMP.isd
IMP_RMF_HAS_IMP_CGAL = _IMP_rmf.IMP_RMF_HAS_IMP_CGAL

IMP_RMF_HAS_IMP_EM = _IMP_rmf.IMP_RMF_HAS_IMP_EM

IMP_RMF_HAS_IMP_SAXS = _IMP_rmf.IMP_RMF_HAS_IMP_SAXS

IMP_RMF_HAS_IMP_SCORE_FUNCTOR = _IMP_rmf.IMP_RMF_HAS_IMP_SCORE_FUNCTOR

IMP_RMF_HAS_IMP_STATISTICS = _IMP_rmf.IMP_RMF_HAS_IMP_STATISTICS

IMP_RMF_HAS_BOOST_FILESYSTEM = _IMP_rmf.IMP_RMF_HAS_BOOST_FILESYSTEM

IMP_RMF_HAS_BOOST_PROGRAMOPTIONS = _IMP_rmf.IMP_RMF_HAS_BOOST_PROGRAMOPTIONS

IMP_RMF_HAS_BOOST_RANDOM = _IMP_rmf.IMP_RMF_HAS_BOOST_RANDOM

IMP_RMF_HAS_BOOST_REGEX = _IMP_rmf.IMP_RMF_HAS_BOOST_REGEX

IMP_RMF_HAS_BOOST_SYSTEM = _IMP_rmf.IMP_RMF_HAS_BOOST_SYSTEM

IMP_RMF_HAS_CGAL = _IMP_rmf.IMP_RMF_HAS_CGAL

IMP_RMF_HAS_HDF5 = _IMP_rmf.IMP_RMF_HAS_HDF5

IMP_RMF_HAS_NUMPY = _IMP_rmf.IMP_RMF_HAS_NUMPY

IMP_RMF_HAS_PYTHON_IHM = _IMP_rmf.IMP_RMF_HAS_PYTHON_IHM

IMPRMF_SHOW_WARNINGS = _IMP_rmf.IMPRMF_SHOW_WARNINGS

import RMF

_object_types.append("SaveOptimizerState")


def _object_cast_to_SaveOptimizerState(o):
    r"""_object_cast_to_SaveOptimizerState(Object o) -> SaveOptimizerState"""
    return _IMP_rmf._object_cast_to_SaveOptimizerState(o)

_object_types.append("LoadLink")


def _object_cast_to_LoadLink(o):
    r"""_object_cast_to_LoadLink(Object o) -> LoadLink"""
    return _IMP_rmf._object_cast_to_LoadLink(o)

_object_types.append("SaveLink")


def _object_cast_to_SaveLink(o):
    r"""_object_cast_to_SaveLink(Object o) -> SaveLink"""
    return _IMP_rmf._object_cast_to_SaveLink(o)

_object_types.append("HierarchySaveLink")


def _object_cast_to_HierarchySaveLink(o):
    r"""_object_cast_to_HierarchySaveLink(Object o) -> HierarchySaveLink"""
    return _IMP_rmf._object_cast_to_HierarchySaveLink(o)

_object_types.append("HierarchyLoadLink")


def _object_cast_to_HierarchyLoadLink(o):
    r"""_object_cast_to_HierarchyLoadLink(Object o) -> HierarchyLoadLink"""
    return _IMP_rmf._object_cast_to_HierarchyLoadLink(o)
class LoadLink(IMP.Object):
    r"""Proxy of C++ IMP::rmf::LoadLink class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def do_load(self, fh):
        r"""do_load(LoadLink self, FileConstHandle fh)"""
        return _IMP_rmf.LoadLink_do_load(self, fh)

    def __init__(self, name):
        r"""__init__(LoadLink self, std::string name) -> LoadLink"""
        if self.__class__ == LoadLink:
            _self = None
        else:
            _self = self
        _IMP_rmf.LoadLink_swiginit(self, _IMP_rmf.new_LoadLink(_self, name))

        if self.__class__ != LoadLink:
            _director_objects.register(self)




    def load(self, fh):
        r"""load(LoadLink self, FileConstHandle fh)"""
        return _IMP_rmf.LoadLink_load(self, fh)
    __swig_destroy__ = _IMP_rmf.delete_LoadLink

    def __str__(self):
        r"""__str__(LoadLink self) -> std::string"""
        return _IMP_rmf.LoadLink___str__(self)

    def __repr__(self):
        r"""__repr__(LoadLink self) -> std::string"""
        return _IMP_rmf.LoadLink___repr__(self)

    @staticmethod
    def get_from(o):
       return _object_cast_to_LoadLink(o)


    def get_type_name(self):
        return self.__class__.__name__
    def do_show(self, out):
        pass
    def get_version_info(self):
        if"IMP::rmf" == "IMP":
          return VersionInfo(self.__module__,
                             __import__(self.__module__).get_module_version())
        else:
          return IMP.VersionInfo(self.__module__,
                             __import__(self.__module__).get_module_version())
    @staticmethod
    def get_from(o):
       return _object_cast_to_LoadLink(o)

    def __disown__(self):
        self.this.disown()
        _IMP_rmf.disown_LoadLink(self)
        return weakref.proxy(self)

    def do_destroy(self):
        r"""do_destroy(LoadLink self)"""
        return _IMP_rmf.LoadLink_do_destroy(self)

# Register LoadLink in _IMP_rmf:
_IMP_rmf.LoadLink_swigregister(LoadLink)

class SaveLink(IMP.Object):
    r"""Proxy of C++ IMP::rmf::SaveLink class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def do_save(self, hf):
        r"""do_save(SaveLink self, FileHandle hf)"""
        return _IMP_rmf.SaveLink_do_save(self, hf)

    def __init__(self, name):
        r"""__init__(SaveLink self, std::string name) -> SaveLink"""
        if self.__class__ == SaveLink:
            _self = None
        else:
            _self = self
        _IMP_rmf.SaveLink_swiginit(self, _IMP_rmf.new_SaveLink(_self, name))

        if self.__class__ != SaveLink:
            _director_objects.register(self)




    def save(self, fh):
        r"""save(SaveLink self, FileHandle fh)"""
        return _IMP_rmf.SaveLink_save(self, fh)
    __swig_destroy__ = _IMP_rmf.delete_SaveLink

    def __str__(self):
        r"""__str__(SaveLink self) -> std::string"""
        return _IMP_rmf.SaveLink___str__(self)

    def __repr__(self):
        r"""__repr__(SaveLink self) -> std::string"""
        return _IMP_rmf.SaveLink___repr__(self)

    @staticmethod
    def get_from(o):
       return _object_cast_to_SaveLink(o)


    def get_type_name(self):
        return self.__class__.__name__
    def do_show(self, out):
        pass
    def get_version_info(self):
        if"IMP::rmf" == "IMP":
          return VersionInfo(self.__module__,
                             __import__(self.__module__).get_module_version())
        else:
          return IMP.VersionInfo(self.__module__,
                             __import__(self.__module__).get_module_version())
    @staticmethod
    def get_from(o):
       return _object_cast_to_SaveLink(o)

    def __disown__(self):
        self.this.disown()
        _IMP_rmf.disown_SaveLink(self)
        return weakref.proxy(self)

    def do_destroy(self):
        r"""do_destroy(SaveLink self)"""
        return _IMP_rmf.SaveLink_do_destroy(self)

# Register SaveLink in _IMP_rmf:
_IMP_rmf.SaveLink_swigregister(SaveLink)


def add_hierarchies(*args):
    r"""
    add_hierarchies(FileHandle fh, IMP::atom::Hierarchies const & hs)
    add_hierarchies(NodeHandle fh, IMP::atom::Hierarchies const & hs)
    """
    return _IMP_rmf.add_hierarchies(*args)

def add_hierarchy(fh, hs):
    r"""add_hierarchy(FileHandle fh, Hierarchy hs)"""
    return _IMP_rmf.add_hierarchy(fh, hs)

def create_hierarchies(fh, m):
    r"""create_hierarchies(FileConstHandle fh, Model m) -> IMP::atom::Hierarchies"""
    return _IMP_rmf.create_hierarchies(fh, m)

def link_hierarchies(fh, hs):
    r"""link_hierarchies(FileConstHandle fh, IMP::atom::Hierarchies const & hs)"""
    return _IMP_rmf.link_hierarchies(fh, hs)

def add_restraints(*args):
    r"""
    add_restraints(FileHandle fh, IMP::Restraints const & hs)
    add_restraints(NodeHandle fh, IMP::Restraints const & hs)
    """
    return _IMP_rmf.add_restraints(*args)

def add_restraint(fh, hs):
    r"""add_restraint(FileHandle fh, Restraint hs)"""
    return _IMP_rmf.add_restraint(fh, hs)

def create_restraints(fh, m):
    r"""create_restraints(FileConstHandle fh, Model m) -> IMP::Restraints"""
    return _IMP_rmf.create_restraints(fh, m)

def link_restraints(fh, hs):
    r"""link_restraints(FileConstHandle fh, IMP::Restraints const & hs)"""
    return _IMP_rmf.link_restraints(fh, hs)

def add_restraints_as_bonds(fh, rs):
    r"""add_restraints_as_bonds(FileHandle fh, IMP::Restraints const & rs)"""
    return _IMP_rmf.add_restraints_as_bonds(fh, rs)

def set_maximum_number_of_terms(fh, num):
    r"""set_maximum_number_of_terms(FileHandle fh, unsigned int num)"""
    return _IMP_rmf.set_maximum_number_of_terms(fh, num)

def add_particles(*args):
    r"""
    add_particles(FileHandle fh, IMP::ParticlesTemp const & hs)
    add_particles(NodeHandle fh, IMP::ParticlesTemp const & hs)
    """
    return _IMP_rmf.add_particles(*args)

def add_particle(fh, hs):
    r"""add_particle(FileHandle fh, Particle hs)"""
    return _IMP_rmf.add_particle(fh, hs)

def create_particles(fh, m):
    r"""create_particles(FileConstHandle fh, Model m) -> IMP::ParticlesTemp"""
    return _IMP_rmf.create_particles(fh, m)

def link_particles(fh, hs):
    r"""link_particles(FileConstHandle fh, IMP::ParticlesTemp const & hs)"""
    return _IMP_rmf.link_particles(fh, hs)

def add_geometries(*args):
    r"""
    add_geometries(FileHandle file, IMP::display::GeometriesTemp const & r)
    add_geometries(NodeHandle parent, IMP::display::GeometriesTemp const & r)
    """
    return _IMP_rmf.add_geometries(*args)

def add_static_geometries(parent, r):
    r"""add_static_geometries(FileHandle parent, IMP::display::GeometriesTemp const & r)"""
    return _IMP_rmf.add_static_geometries(parent, r)

def add_geometry(file, r):
    r"""add_geometry(FileHandle file, Geometry r)"""
    return _IMP_rmf.add_geometry(file, r)

def create_geometries(parent):
    r"""create_geometries(FileConstHandle parent) -> IMP::display::Geometries"""
    return _IMP_rmf.create_geometries(parent)

def link_geometries(parent, r):
    r"""link_geometries(FileConstHandle parent, IMP::display::GeometriesTemp const & r)"""
    return _IMP_rmf.link_geometries(parent, r)

def set_association(nh, o, overwrite=False):
    r"""set_association(NodeConstHandle nh, Object o, bool overwrite=False)"""
    return _IMP_rmf.set_association(nh, o, overwrite)

def get_node_from_association(*args):
    r"""
    get_node_from_association(FileConstHandle nh, Object oi) -> NodeConstHandle
    get_node_from_association(FileHandle nh, Object oi) -> NodeHandle
    """
    return _IMP_rmf.get_node_from_association(*args)

def get_has_associated_node(nh, oi):
    r"""get_has_associated_node(FileConstHandle nh, Object oi) -> bool"""
    return _IMP_rmf.get_has_associated_node(nh, oi)
class SaveOptimizerState(IMP.OptimizerState):
    r"""Proxy of C++ IMP::rmf::SaveOptimizerState class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, m, fh):
        r"""__init__(SaveOptimizerState self, Model m, FileHandle fh) -> SaveOptimizerState"""
        _IMP_rmf.SaveOptimizerState_swiginit(self, _IMP_rmf.new_SaveOptimizerState(m, fh))

    def remove_hierarchy(self, d):
        r"""remove_hierarchy(SaveOptimizerState self, Particle d)"""
        return _IMP_rmf.SaveOptimizerState_remove_hierarchy(self, d)

    def remove_hierarchies(self, d):
        r"""remove_hierarchies(SaveOptimizerState self, IMP::Particles const & d)"""
        return _IMP_rmf.SaveOptimizerState_remove_hierarchies(self, d)

    def set_hierarchies(self, ps):
        r"""set_hierarchies(SaveOptimizerState self, IMP::Particles const & ps)"""
        return _IMP_rmf.SaveOptimizerState_set_hierarchies(self, ps)

    def set_hierarchies_order(self, objs):
        r"""set_hierarchies_order(SaveOptimizerState self, IMP::Particles const & objs)"""
        return _IMP_rmf.SaveOptimizerState_set_hierarchies_order(self, objs)

    def add_hierarchy(self, obj):
        r"""add_hierarchy(SaveOptimizerState self, Particle obj) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_add_hierarchy(self, obj)

    def add_hierarchies(self, objs):
        r"""add_hierarchies(SaveOptimizerState self, IMP::Particles const & objs)"""
        return _IMP_rmf.SaveOptimizerState_add_hierarchies(self, objs)

    def clear_hierarchies(self):
        r"""clear_hierarchies(SaveOptimizerState self)"""
        return _IMP_rmf.SaveOptimizerState_clear_hierarchies(self)

    def get_number_of_hierarchies(self):
        r"""get_number_of_hierarchies(SaveOptimizerState self) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_get_number_of_hierarchies(self)

    def get_has_hierarchies(self):
        r"""get_has_hierarchies(SaveOptimizerState self) -> bool"""
        return _IMP_rmf.SaveOptimizerState_get_has_hierarchies(self)

    def get_hierarchy(self, i):
        r"""get_hierarchy(SaveOptimizerState self, unsigned int i) -> Particle"""
        return _IMP_rmf.SaveOptimizerState_get_hierarchy(self, i)

    def get_hierarchies(self):
        r"""get_hierarchies(SaveOptimizerState self) -> IMP::Particles"""
        return _IMP_rmf.SaveOptimizerState_get_hierarchies(self)

    def reserve_hierarchies(self, sz):
        r"""reserve_hierarchies(SaveOptimizerState self, unsigned int sz)"""
        return _IMP_rmf.SaveOptimizerState_reserve_hierarchies(self, sz)

    def remove_restraint(self, d):
        r"""remove_restraint(SaveOptimizerState self, Restraint d)"""
        return _IMP_rmf.SaveOptimizerState_remove_restraint(self, d)

    def remove_restraints(self, d):
        r"""remove_restraints(SaveOptimizerState self, IMP::Restraints const & d)"""
        return _IMP_rmf.SaveOptimizerState_remove_restraints(self, d)

    def set_restraints(self, ps):
        r"""set_restraints(SaveOptimizerState self, IMP::Restraints const & ps)"""
        return _IMP_rmf.SaveOptimizerState_set_restraints(self, ps)

    def set_restraints_order(self, objs):
        r"""set_restraints_order(SaveOptimizerState self, IMP::Restraints const & objs)"""
        return _IMP_rmf.SaveOptimizerState_set_restraints_order(self, objs)

    def add_restraint(self, obj):
        r"""add_restraint(SaveOptimizerState self, Restraint obj) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_add_restraint(self, obj)

    def add_restraints(self, objs):
        r"""add_restraints(SaveOptimizerState self, IMP::Restraints const & objs)"""
        return _IMP_rmf.SaveOptimizerState_add_restraints(self, objs)

    def clear_restraints(self):
        r"""clear_restraints(SaveOptimizerState self)"""
        return _IMP_rmf.SaveOptimizerState_clear_restraints(self)

    def get_number_of_restraints(self):
        r"""get_number_of_restraints(SaveOptimizerState self) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_get_number_of_restraints(self)

    def get_has_restraints(self):
        r"""get_has_restraints(SaveOptimizerState self) -> bool"""
        return _IMP_rmf.SaveOptimizerState_get_has_restraints(self)

    def get_restraint(self, i):
        r"""get_restraint(SaveOptimizerState self, unsigned int i) -> Restraint"""
        return _IMP_rmf.SaveOptimizerState_get_restraint(self, i)

    def get_restraints(self):
        r"""get_restraints(SaveOptimizerState self) -> IMP::Restraints"""
        return _IMP_rmf.SaveOptimizerState_get_restraints(self)

    def reserve_restraints(self, sz):
        r"""reserve_restraints(SaveOptimizerState self, unsigned int sz)"""
        return _IMP_rmf.SaveOptimizerState_reserve_restraints(self, sz)

    def remove_particle(self, d):
        r"""remove_particle(SaveOptimizerState self, Particle d)"""
        return _IMP_rmf.SaveOptimizerState_remove_particle(self, d)

    def remove_particles(self, d):
        r"""remove_particles(SaveOptimizerState self, IMP::Particles const & d)"""
        return _IMP_rmf.SaveOptimizerState_remove_particles(self, d)

    def set_particles(self, ps):
        r"""set_particles(SaveOptimizerState self, IMP::Particles const & ps)"""
        return _IMP_rmf.SaveOptimizerState_set_particles(self, ps)

    def set_particles_order(self, objs):
        r"""set_particles_order(SaveOptimizerState self, IMP::Particles const & objs)"""
        return _IMP_rmf.SaveOptimizerState_set_particles_order(self, objs)

    def add_particle(self, obj):
        r"""add_particle(SaveOptimizerState self, Particle obj) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_add_particle(self, obj)

    def add_particles(self, objs):
        r"""add_particles(SaveOptimizerState self, IMP::Particles const & objs)"""
        return _IMP_rmf.SaveOptimizerState_add_particles(self, objs)

    def clear_particles(self):
        r"""clear_particles(SaveOptimizerState self)"""
        return _IMP_rmf.SaveOptimizerState_clear_particles(self)

    def get_number_of_particles(self):
        r"""get_number_of_particles(SaveOptimizerState self) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_get_number_of_particles(self)

    def get_has_particles(self):
        r"""get_has_particles(SaveOptimizerState self) -> bool"""
        return _IMP_rmf.SaveOptimizerState_get_has_particles(self)

    def get_particle(self, i):
        r"""get_particle(SaveOptimizerState self, unsigned int i) -> Particle"""
        return _IMP_rmf.SaveOptimizerState_get_particle(self, i)

    def get_particles(self):
        r"""get_particles(SaveOptimizerState self) -> IMP::Particles"""
        return _IMP_rmf.SaveOptimizerState_get_particles(self)

    def reserve_particles(self, sz):
        r"""reserve_particles(SaveOptimizerState self, unsigned int sz)"""
        return _IMP_rmf.SaveOptimizerState_reserve_particles(self, sz)

    def remove_geometry(self, d):
        r"""remove_geometry(SaveOptimizerState self, Geometry d)"""
        return _IMP_rmf.SaveOptimizerState_remove_geometry(self, d)

    def remove_geometries(self, d):
        r"""remove_geometries(SaveOptimizerState self, IMP::display::Geometries const & d)"""
        return _IMP_rmf.SaveOptimizerState_remove_geometries(self, d)

    def set_geometries(self, ps):
        r"""set_geometries(SaveOptimizerState self, IMP::display::Geometries const & ps)"""
        return _IMP_rmf.SaveOptimizerState_set_geometries(self, ps)

    def set_geometries_order(self, objs):
        r"""set_geometries_order(SaveOptimizerState self, IMP::display::Geometries const & objs)"""
        return _IMP_rmf.SaveOptimizerState_set_geometries_order(self, objs)

    def add_geometry(self, obj):
        r"""add_geometry(SaveOptimizerState self, Geometry obj) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_add_geometry(self, obj)

    def add_geometries(self, objs):
        r"""add_geometries(SaveOptimizerState self, IMP::display::Geometries const & objs)"""
        return _IMP_rmf.SaveOptimizerState_add_geometries(self, objs)

    def clear_geometries(self):
        r"""clear_geometries(SaveOptimizerState self)"""
        return _IMP_rmf.SaveOptimizerState_clear_geometries(self)

    def get_number_of_geometries(self):
        r"""get_number_of_geometries(SaveOptimizerState self) -> unsigned int"""
        return _IMP_rmf.SaveOptimizerState_get_number_of_geometries(self)

    def get_has_geometries(self):
        r"""get_has_geometries(SaveOptimizerState self) -> bool"""
        return _IMP_rmf.SaveOptimizerState_get_has_geometries(self)

    def get_geometry(self, i):
        r"""get_geometry(SaveOptimizerState self, unsigned int i) -> Geometry"""
        return _IMP_rmf.SaveOptimizerState_get_geometry(self, i)

    def get_geometries(self):
        r"""get_geometries(SaveOptimizerState self) -> IMP::display::Geometries"""
        return _IMP_rmf.SaveOptimizerState_get_geometries(self)

    def reserve_geometries(self, sz):
        r"""reserve_geometries(SaveOptimizerState self, unsigned int sz)"""
        return _IMP_rmf.SaveOptimizerState_reserve_geometries(self, sz)

    def set_simulator(self, sim):
        r"""set_simulator(SaveOptimizerState self, Simulator sim)"""
        return _IMP_rmf.SaveOptimizerState_set_simulator(self, sim)

    def update_always(self, *args):
        r"""
        update_always(SaveOptimizerState self, std::string name)
        update_always(SaveOptimizerState self)
        """
        return _IMP_rmf.SaveOptimizerState_update_always(self, *args)

    def do_update(self, call_num):
        r"""do_update(SaveOptimizerState self, unsigned int call_num)"""
        return _IMP_rmf.SaveOptimizerState_do_update(self, call_num)

    def get_version_info(self):
        r"""get_version_info(SaveOptimizerState self) -> VersionInfo"""
        return _IMP_rmf.SaveOptimizerState_get_version_info(self)

    def __str__(self):
        r"""__str__(SaveOptimizerState self) -> std::string"""
        return _IMP_rmf.SaveOptimizerState___str__(self)

    def __repr__(self):
        r"""__repr__(SaveOptimizerState self) -> std::string"""
        return _IMP_rmf.SaveOptimizerState___repr__(self)

    @staticmethod
    def get_from(o):
       return _object_cast_to_SaveOptimizerState(o)


# Register SaveOptimizerState in _IMP_rmf:
_IMP_rmf.SaveOptimizerState_swigregister(SaveOptimizerState)


def load_frame(file, frame):
    r"""load_frame(FileConstHandle file, FrameID frame)"""
    return _IMP_rmf.load_frame(file, frame)

def save_frame(*args):
    r"""save_frame(FileHandle file, std::string name="") -> FrameID"""
    return _IMP_rmf.save_frame(*args)
class _ParticleSaveLink(SaveLink):
    r"""Proxy of C++ IMP::rmf::SimpleSaveLink< IMP::Particle > class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr

    def add(self, parent, os):
        r"""add(_ParticleSaveLink self, NodeHandle parent, IMP::Vector< Pointer< IMP::Particle > > const & os)"""
        return _IMP_rmf._ParticleSaveLink_add(self, parent, os)
    __swig_destroy__ = _IMP_rmf.delete__ParticleSaveLink

# Register _ParticleSaveLink in _IMP_rmf:
_IMP_rmf._ParticleSaveLink_swigregister(_ParticleSaveLink)

class _ParticleLoadLink(LoadLink):
    r"""Proxy of C++ IMP::rmf::SimpleLoadLink< IMP::Particle > class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr

    def create(self, *args):
        r"""
        create(_ParticleLoadLink self, NodeConstHandle rt) -> IMP::Vector< Pointer< IMP::Particle > >
        create(_ParticleLoadLink self, NodeConstHandle rt, Model m) -> IMP::Vector< Pointer< IMP::Particle > >
        """
        return _IMP_rmf._ParticleLoadLink_create(self, *args)

    def link(self, rt, ps):
        r"""link(_ParticleLoadLink self, NodeConstHandle rt, IMP::Vector< Pointer< IMP::Particle > > const & ps)"""
        return _IMP_rmf._ParticleLoadLink_link(self, rt, ps)
    __swig_destroy__ = _IMP_rmf.delete__ParticleLoadLink

# Register _ParticleLoadLink in _IMP_rmf:
_IMP_rmf._ParticleLoadLink_swigregister(_ParticleLoadLink)

class HierarchyLoadLink(_ParticleLoadLink):
    r"""Proxy of C++ IMP::rmf::HierarchyLoadLink class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, fh):
        r"""__init__(HierarchyLoadLink self, FileConstHandle fh) -> HierarchyLoadLink"""
        _IMP_rmf.HierarchyLoadLink_swiginit(self, _IMP_rmf.new_HierarchyLoadLink(fh))

    @staticmethod
    def get_name():
        r"""get_name() -> char const *"""
        return _IMP_rmf.HierarchyLoadLink_get_name()

    def get_version_info(self):
        r"""get_version_info(HierarchyLoadLink self) -> VersionInfo"""
        return _IMP_rmf.HierarchyLoadLink_get_version_info(self)

    def __str__(self):
        r"""__str__(HierarchyLoadLink self) -> std::string"""
        return _IMP_rmf.HierarchyLoadLink___str__(self)

    def __repr__(self):
        r"""__repr__(HierarchyLoadLink self) -> std::string"""
        return _IMP_rmf.HierarchyLoadLink___repr__(self)

    @staticmethod
    def get_from(o):
       return _object_cast_to_HierarchyLoadLink(o)


# Register HierarchyLoadLink in _IMP_rmf:
_IMP_rmf.HierarchyLoadLink_swigregister(HierarchyLoadLink)

def HierarchyLoadLink_get_name():
    r"""HierarchyLoadLink_get_name() -> char const *"""
    return _IMP_rmf.HierarchyLoadLink_get_name()

class HierarchySaveLink(_ParticleSaveLink):
    r"""Proxy of C++ IMP::rmf::HierarchySaveLink class."""

    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, fh):
        r"""__init__(HierarchySaveLink self, FileHandle fh) -> HierarchySaveLink"""
        _IMP_rmf.HierarchySaveLink_swiginit(self, _IMP_rmf.new_HierarchySaveLink(fh))

    @staticmethod
    def get_name():
        r"""get_name() -> char const *"""
        return _IMP_rmf.HierarchySaveLink_get_name()

    def get_version_info(self):
        r"""get_version_info(HierarchySaveLink self) -> VersionInfo"""
        return _IMP_rmf.HierarchySaveLink_get_version_info(self)

    def __str__(self):
        r"""__str__(HierarchySaveLink self) -> std::string"""
        return _IMP_rmf.HierarchySaveLink___str__(self)

    def __repr__(self):
        r"""__repr__(HierarchySaveLink self) -> std::string"""
        return _IMP_rmf.HierarchySaveLink___repr__(self)

    @staticmethod
    def get_from(o):
       return _object_cast_to_HierarchySaveLink(o)


# Register HierarchySaveLink in _IMP_rmf:
_IMP_rmf.HierarchySaveLink_swigregister(HierarchySaveLink)

def HierarchySaveLink_get_name():
    r"""HierarchySaveLink_get_name() -> char const *"""
    return _IMP_rmf.HierarchySaveLink_get_name()


suffixes = [".rmf", ".rmfz"]


def get_module_version():
    r"""get_module_version() -> std::string const"""
    return _IMP_rmf.get_module_version()

def get_example_path(fname):
    r"""get_example_path(std::string fname) -> std::string"""
    return _IMP_rmf.get_example_path(fname)

def get_data_path(fname):
    r"""get_data_path(std::string fname) -> std::string"""
    return _IMP_rmf.get_data_path(fname)

from . import _version_check
_version_check.check_version(get_module_version())
__version__ = get_module_version()






"""@namespace IMP.EMageFit.buildxlinks
   Utility functions to handle cross links.
"""

import IMP
import IMP.atom as atom
import IMP.algebra as alg
import IMP.core as core
import sys
import logging
from IMP.EMageFit.argminmax import keymax
log = logging.getLogger("buildxlinks")

import networkx as nx

try:
    set = set
except NameError:
    from sets import Set as set


class Xlink:

    """
        Class defining a cross-link
    """

    def __init__(self, id1, chain1, residue1,
                 id2, chain2, residue2,
                 distance):
        """
            Initialize the class
            @param[in] id1 Id of the first component of the cross-link
            @param[in] chain1 Chain of the first component
            @param[in] residue1 Residue cross-linked in the first component
            @param[in] id2   Id of the second component of the cross-link
            @param[in] chain2   chain1 Chain of the second component
            @param[in] residue2   Residue cross-linked in the second component
            @param[in] distance Maximum distance
        """
        self.first_id = id1
        self.first_chain = chain1
        self.first_residue = residue1
        self.second_id = id2
        self.second_chain = chain2
        self.second_residue = residue2
        self.distance = distance

    def clone(self):
        xl = Xlink(self.first_id, self.first_chain, self.first_residue,
                   self.second_id, self.second_chain, self.second_residue,
                   self.distance)
        return xl

    def __eq__(self, other):
        if self.first_id != other.first_id or \
           self.second_id != other.second_id or \
           self.first_chain != other.first_chain or \
           self.second_chain != other.second_chain or \
           self.first_residue != other.first_residue or \
           self.second_residue != other.second_residue or \
           abs(other.distance - self.distance) > 0.01:
            return False
        return True

    def swap(self):
        """
            swaps the order of the residues in the restraint
        """
        self.first_id, self.second_id = self.second_id, self.first_id
        self.first_residue, self.second_residue = \
            self.second_residue, self.first_residue
        self.first_chain, self.second_chain = \
            self.second_chain, self.first_chain

    def show(self):
        s = "Cross Link: %s %s %d - %s %s %d. Distance %f" % (self.first_id,
                                                              self.first_chain, self.first_residue,
                                                              self.second_id, self.second_chain,
                                                              self.second_residue, self.distance)
        return s

    def get_name(self):
        """
            Generate a unique name for the restraint.
            @note: The name cannot start with a number, upsets sqlite
        """
        name = "cl_%s_%s%d_%s_%s%d" % (self.first_id, self.first_chain,
                                       self.first_residue, self.second_id,
                                       self.second_chain, self.second_residue)
        return name


def build_xlinks_graph(xlinks_dict):
    """
        Build a set of nodes and edges from the set of crosslinking
        restraints.
        @param xlinks_dict a XlinksDict class

    """
    subunits = set()
    edges = []
    for key in xlinks_dict.keys():
        subunits.add(key[0])
        subunits.add(key[1])
        edge = sorted([key[0], key[1]])
        if edge not in edges:
            edges.append(edge)
    log.debug("Subunits %s", subunits)
    log.debug("Edges %s", edges)
    return subunits, edges


class XlinksDict(dict):

    """
        Description of crosslinking restraints as a python
        dictionary.
        The keys are a pair with the ids of the cross-linked subunits.
        Note: The pairs are considered in alphabetic order
    """

    def add(self, xlink):
        """
            Add a xlink. It is ensured that the id of the first element is
            is lower that the second
        """
        if xlink.second_id < xlink.first_id:
            xlink.swap()
        key = (xlink.first_id, xlink.second_id)
        if key not in self.keys():
            self[key] = []
        self[key].append(xlink)

    def get_xlinks_for_pair(self, pair_ids):
        """
            @param pair_ids Ids fo the subunits that are cross-linked
        """
        try:
            xlinks_list = self[pair_ids]
            ys = [xl.clone() for xl in xlinks_list]

        except KeyError:
            try:
                xlinks_list = self[(pair_ids[1], pair_ids[0])]
                ys = [xl.clone() for xl in xlinks_list]
                # Ensure that the first element of the pair corresponds to the
                # first id in the xlink
                for y in ys:
                    y.swap()
            except KeyError as e:
                raise e
        return ys

        """
        try:
            return self[pair_ids]
        except KeyError:
            print pair_ids,"NOT FOUND, swapping to",(pair_ids[1], pair_ids[0])
            # If not found, invert the pair
            xlinks_list = self[(pair_ids[1], pair_ids[0])]
            # swap the xlinks, so the first element of the pair corresponds to the
            # first id in the xlink
            for xl in xlinks_list:
                xl.swap()
            for xl in xlinks_list:
                print xl.show()
            return xlinks_list
        """


class DockOrder (object):

    """
        Compute the order of the docking experiments. The order is derived
        from the cross-linking restraints:
        1) The subunit with the highest number of cross-links with others
           works as the first receptor (A)
        2) All the subunits cross-linked with A are docked into A.
        3) The next receptor (B) is the subunit that has the highest number of
            cross-links after A.
        4) All the subunits cross-linked to B are docked into B (except A)
        5) The procedure is repeated until there are no more cross-links
    """

    def __init__(self,):
        self.G = nx.Graph()

    def set_components_and_connections(self, subunits, edges):
        """
            Instead of setting the xlink restraints, init the graph directly
            Example of what to pass with the structure in 3sfd
            G.add_nodes_from(["A", "B", "C", "D"])
            G.add_edges_from([("B","C"), ("B","A"), ("B","D"), ("C","D")])
            ("B","C") means that there are cross-links between B and C
        """
        self.G.add_nodes_from(subunits)
        self.G.add_edges_from(edges)

    def set_xlinks(self, xlinks_dict):
        """
            Sets the xlinks used for computing the docking order
            @param xlinks_dict XlinksDict class
        """
        subunits, edges = build_xlinks_graph(xlinks_dict)
        self.set_components_and_connections(subunits, edges)

    def get_docking_order(self):
        """ return the order to dock components from the cross links """
        docking_pairs = []
        degs = self.G.degree(self.G.nodes())
        log.debug("Degrees: %s", degs)
        sorted_degrees = sorted(((v,k) for (k,v) in dict(degs).items()),
                                reverse=True)

        receptors_considered = []
        for degree, node in sorted_degrees:
            for n in self.G.neighbors(node):
                if not n in receptors_considered:
                    docking_pairs.append((node, n))
            receptors_considered.append(node)
        log.info("The suggested order for the docking pairs is %s",
                 docking_pairs)
        return docking_pairs


class InitialDockingFromXlinks:

    """
        Puts two subunits together using the Xlinkins restraints. The solutions
        offered by this class are just an initial position of the components
        to be fed to HEX
    """

    def clear_xlinks(self):
        self.xlinks_list = []

    def set_xlinks(self, xlinks_list):
        """
            Sets the xlinks used for the computation fo the initial rough
            docking solution
            @param xlinks_list A list of Xlink classes
            residue1 belongs to the receptor and residue2 belongs to the ligand
        """
        self.xlinks_list = xlinks_list

    def set_pdbs(self, fn_receptor, fn_ligand):
        """
            Set the name of the PDB files of the receptor and the ligand
            @param fn_receptor
            @param fn_ligand
        """
        sel = atom.ATOMPDBSelector()
        self.m_receptor = IMP.Model()
        self.h_receptor = atom.read_pdb(fn_receptor, self.m_receptor, sel)
        self.m_ligand = IMP.Model()
        self.h_ligand = atom.read_pdb(fn_ligand, self.m_ligand, sel)

    def set_hierarchies(self, h_receptor, h_ligand):
        """
            Set the hierarchies (atom.Hierarchy objects) for the receptor and
            the ligand
            @param h_receptor
            @param h_ligand
        """
        self.h_receptor = h_receptor
        self.h_ligand = h_ligand

    def set_rigid_bodies(self, rb_receptor, rb_ligand):
        """
            Sets the rigid bodies (core.RigidBody objects) for the receptor and
            the ligand
            @param rb_receptor
            @param rb_ligand
        """
        self.rb_receptor = rb_receptor
        self.rb_ligand = rb_ligand

    def move_ligand(self):
        """
            Movest the ligand close to the receptor based on the xlinks
            provided by set_xlinks()
        """
        log.debug("Moving ligand close to the receptor using the xlinks")
        n = len(self.xlinks_list)
        if n == 1:
            self.move_one_xlink()
        elif n >= 2:
            self.move_xlinks()

    def get_residue_particle(self, h, ch, res):
        """
            Get the particle representing a residue in a hierarchy
            @param h atom.Hierarchy containing the residue
            @param ch The chain id
            @param res index of the residue
        """
        s = IMP.atom.Selection(h, chain=ch, residue_index=res)
        return s.get_selected_particles()[0]

    def get_residue_coordinates(self, h, ch, res):
        """
            Get the coordinates for a residue in a molecular hierarchy
            @param h atom.Hierarchy object
            @param ch The chain id
            @param res Residue index
        """
        p = self.get_residue_particle(h, ch, res)
        return core.XYZ(p).get_coordinates()

    def write_ligand(self, fn):
        """
            Write a pdb file the coordinates of the ligand
            @param fn
        """
        atom.write_pdb(self.h_ligand, fn)

    def move_one_xlink(self):
        """
            Put the residues in a random distance between 0 and the maximum
            cross-linkin distance
        """
        xl = self.xlinks_list[0]
        center = self.get_residue_coordinates(self.h_receptor, xl.first_chain,
                                              xl.first_residue)
        sph = alg.Sphere3D(center, xl.distance)
        v = alg.get_random_vector_in(sph)
        ref = self.rb_ligand.get_reference_frame()
        coords = ref.get_transformation_to().get_translation()
        R = ref.get_transformation_to().get_rotation()
        lig = self.get_residue_coordinates(self.h_ligand, xl.second_chain,
                                           xl.second_residue)
        log.debug("Ligand residue before moving %s", lig)
        displacement = v - lig
        T = alg.Transformation3D(R, coords + displacement)
        self.rb_ligand.set_reference_frame(alg.ReferenceFrame3D(T))
        new_coords = self.get_residue_coordinates(
            self.h_ligand, xl.second_chain,
            xl.second_residue)
        log.debug("ligand after moving %s", new_coords)

    def move_xlinks(self, ):
        """
            Function equivalent to move_one_xlink() for the case where there
            are more than one cross-link restraints available.
            Puts the ligand residues as close as possible to the receptor
            residues
        """
        rec_coords = []
        lig_coords = []
        for xl in self.xlinks_list:
            c = self.get_residue_coordinates(self.h_receptor, xl.first_chain,
                                             xl.first_residue)
            rec_coords.append(c)
            c = self.get_residue_coordinates(self.h_ligand, xl.second_chain,
                                             xl.second_residue)
            lig_coords.append(c)
        log.debug("Receptor residues before moving %s", rec_coords)
        log.debug("Ligand residues before moving %s", lig_coords)
        ref = self.rb_ligand.get_reference_frame()
        Tr = alg.get_transformation_aligning_first_to_second(lig_coords,
                                                             rec_coords)
        T = ref.get_transformation_to()
        newT = alg.compose(Tr, T)
        self.rb_ligand.set_reference_frame(alg.ReferenceFrame3D(newT))

        moved_lig_coords = []
        for xl in self.xlinks_list:
            c = self.get_residue_coordinates(self.h_ligand, xl.second_chain,
                                             xl.second_residue)
            moved_lig_coords.append(c)
        log.debug("Ligand residues after moving %s", moved_lig_coords)



"""@namespace IMP.pmi1.metadata
Classes for attaching metadata to PMI objects.

@note All of these classes are deprecated; use classes from the ihm package
      instead.
"""

from __future__ import print_function, division
import IMP
import ihm.location
import ihm.dataset

@IMP.deprecated_object("2.9", "Use ihm.Software instead")
class Software(ihm.Software):
    def __init__(self, name, classification, description, url, type='program',
                 version=None):
        super(Software, self).__init__(name=name, classification=classification,
                        description=description, location=url,
                        type=type, version=version)


@IMP.deprecated_object("2.9", "Use ihm.Citation instead")
class Citation(ihm.Citation):
    pass


@IMP.deprecated_object("2.9", "Use ihm.location.Repository instead")
class Repository(ihm.location.Repository):
    pass


@IMP.deprecated_object("2.9",
                       "Use a subclass of ihm.location.FileLocation instead")
class FileLocation(ihm.location.FileLocation):
    pass


@IMP.deprecated_object("2.9", "Use ihm.dataset.EMMicrographsDataset instead")
class EMMicrographsDataset(ihm.dataset.EMMicrographsDataset):
    def __init__(self, location, number):
        super(EMMicrographsDataset, self).__init__(location)
        self.number = number


@IMP.deprecated_object("2.9", "Use ihm.location.MassIVELocation instead")
class MassIVELocation(ihm.location.MassIVELocation):
    pass


@IMP.deprecated_object("2.9", "Use ihm.dataset.MassSpecDataset instead")
class MassSpecDataset(ihm.dataset.MassSpecDataset):
    pass


@IMP.deprecated_object("2.9", "Use ihm.location.EMDBLocation instead")
class EMDBLocation(ihm.location.EMDBLocation):
    pass


@IMP.deprecated_object("2.9", "Use ihm.location.EMPIARLocation instead")
class EMPIARLocation(ihm.location.EMPIARLocation):
    pass


@IMP.deprecated_object("2.9", "Use ihm.location.SASBDBLocation instead")
class SASBDBLocation(ihm.location.SASBDBLocation):
    pass


@IMP.deprecated_object("2.9", "Use ihm.dataset.EMDensityDataset instead")
class EMDensityDataset(ihm.dataset.EMDensityDataset):
    pass


@IMP.deprecated_object("2.9", "Use ihm.dataset.SASDataset instead")
class SASDataset(ihm.dataset.SASDataset):
    pass


@IMP.deprecated_function("2.9", "Use ihm.location.WorkflowFileLocation instead")
def PythonScript(location):
    location.content_type = ihm.location.WorkflowFileLocation.content_type
    return location


@IMP.deprecated_function("2.9",
                         "Use ihm.location.VisualizationFileLocation instead")
def ChimeraXCommandScript(location):
    location.content_type = ihm.location.VisualizationFileLocation.content_type
    return location



## \example atom/charmm_forcefield.py
# In this example, a PDB file is read in and scored using the CHARMM
# forcefield. For more control over the setup of the forcefield, see
# the 'charmm_forcefield_verbose.py' example.

from __future__ import print_function
import IMP.atom
import IMP.container
import sys

IMP.setup_from_argv(sys.argv, "CHARMM forcefield")

# Create an IMP model and add a heavy atom-only protein from a PDB file
m = IMP.Model()
prot = IMP.atom.read_pdb(IMP.atom.get_example_path("example_protein.pdb"), m,
                         IMP.atom.NonWaterNonHydrogenPDBSelector())

# Read in the CHARMM heavy atom topology and parameter files
ff = IMP.atom.get_heavy_atom_CHARMM_parameters()

# Using the CHARMM libraries, determine the ideal topology (atoms and their
# connectivity) for the PDB file's primary sequence
topology = ff.create_topology(prot)

# Typically this modifies the C and N termini of each chain in the protein by
# applying the CHARMM CTER and NTER patches. Patches can also be manually
# applied at this point, e.g. to add disulfide bridges.
topology.apply_default_patches()

# Make the PDB file conform with the topology; i.e. if it contains extra
# atoms that are not in the CHARMM topology file, remove them; if it is
# missing atoms (e.g. sidechains, hydrogens) that are in the CHARMM topology,
# add them and construct their Cartesian coordinates from internal coordinate
# information.
topology.setup_hierarchy(prot)

# Set up and evaluate the stereochemical part (bonds, angles, dihedrals,
# impropers) of the CHARMM forcefield
r = IMP.atom.CHARMMStereochemistryRestraint(prot, topology)

# Add non-bonded interaction (in this case, Lennard-Jones). This needs to
# know the radii and well depths for each atom, so add them from the forcefield
# (they can also be assigned manually using the XYZR or LennardJones
# decorators):
ff.add_radii(prot)
ff.add_well_depths(prot)

# Get a list of all atoms in the protein, and put it in a container
atoms = IMP.atom.get_by_type(prot, IMP.atom.ATOM_TYPE)
cont = IMP.container.ListSingletonContainer(m, atoms)

# Add a restraint for the Lennard-Jones interaction. This is built from
# a collection of building blocks. First, a ClosePairContainer maintains a list
# of all pairs of Particles that are close. Next, all 1-2, 1-3 and 1-4 pairs
# from the stereochemistry created above are filtered out.
# Then, a LennardJonesPairScore scores a pair of atoms with the Lennard-Jones
# potential. Finally, a PairsRestraint is used which simply applies the
# LennardJonesPairScore to each pair in the ClosePairContainer.
nbl = IMP.container.ClosePairContainer(cont, 4.0)
nbl.add_pair_filter(r.get_pair_filter())

sf = IMP.atom.ForceSwitch(6.0, 7.0)
ps = IMP.atom.LennardJonesPairScore(sf)
restraints = [r, IMP.container.PairsRestraint(ps, nbl)]
scoring_function = IMP.core.RestraintsScoringFunction(restraints)

# it gets awfully slow with internal checks
IMP.set_check_level(IMP.USAGE)
# Finally, evaluate the score of the whole system (without derivatives)
print(scoring_function.evaluate(False))



## \example em/cube.py
# The example creates a simple mrc file that is filled uniformly with 1s.
# In addition, an RMF file is written with a marker at each corner of the
# density. When both files are opened (e.g. in Chimera), the density should
# be centered among the markers. This can be used for testing for
# registration errors when reading and writing density maps.

from __future__ import print_function
import IMP.em
import IMP.display
import IMP.rmf
import RMF
import sys

IMP.setup_from_argv(sys.argv, "cube")

bb = IMP.algebra.BoundingBox3D(IMP.algebra.Vector3D(1, 1, 1),
                               IMP.algebra.Vector3D(11, 11, 11))
dm = IMP.em.create_density_map(bb, 2.5)
dmbb = IMP.em.get_bounding_box(dm)
print(dmbb, bb)

for i in range(0, dm.get_number_of_voxels()):
    dm.set_value(i, 1)

nm = IMP.create_temporary_file_name("cube", ".mrc")
print(nm)
IMP.em.write_map(dm, nm)

nm = IMP.create_temporary_file_name("cube", ".rmf")
print(nm)
w = RMF.create_rmf_file(nm)

for v in IMP.algebra.get_vertices(bb):
    g = IMP.display.SphereGeometry(IMP.algebra.Sphere3D(v, 1))
    IMP.rmf.add_geometry(w, g)
IMP.rmf.save_frame(w, "zero")



"""@namespace IMP.pmi1.topology
Set of python classes to create a multi-state, multi-resolution IMP hierarchy.
* Start by creating a System with `model = IMP.Model(); s = IMP.pmi1.topology.System(model)`. The System will store all the states.
* Then call System.create_state(). You can easily create a multistate system by calling this function multiples times.
* For each State, call State.create_molecule() to add a Molecule (a uniquely named polymer). This function returns the Molecule object which can be passed to various PMI functions.
* Some useful functions to help you set up your Molecules:
 * Access the sequence residues with slicing (Molecule[a:b]) or functions like Molecule.get_atomic_residues() and Molecule.get_non_atomic_residues(). These functions all return python sets for easy set arithmetic using & (and), | (or), - (difference)
 * Molecule.add_structure() to add structural information from a PDB file.
 * Molecule.add_representation() to create a representation unit - here you can choose bead resolutions as well as alternate representations like densities or ideal helices.
 * Molecule.create_clone() lets you set up a molecule with identical representations, just a different chain ID. Use Molecule.create_copy() if you want a molecule with the same sequence but that allows custom representations.
* Once data has been added and representations chosen, call System.build() to create a canonical IMP hierarchy.
* Following hierarchy construction, setup rigid bodies, flexible beads, etc in IMP::pmi1::dof.
* Check your representation with a nice printout: IMP::atom::show_with_representation()
See a [comprehensive example](https://integrativemodeling.org/nightly/doc/ref/pmi_2multiscale_8py-example.html) for using these classes.

Alternatively one can construct the entire topology and degrees of freedom via formatted text file with TopologyReader and IMP::pmi1::macros::BuildSystem(). This is used in the [PMI tutorial](@ref rnapolii_stalk).
Note that this only allows a limited set of the full options available to PMI users (rigid bodies only, fixed resolutions).
"""

from __future__ import print_function
import IMP
import IMP.atom
import IMP.algebra
import IMP.pmi1
import IMP.pmi1.tools
import csv
import os
from collections import defaultdict
from bisect import bisect_left
from math import pi,cos,sin
from operator import itemgetter

def _build_ideal_helix(model, residues, coord_finder):
    """Creates an ideal helix from the specified residue range
    Residues MUST be contiguous.
    This function actually adds them to the TempResidue hierarchy
    """
    created_hiers = []

    # this function creates a CAlpha helix structure (which can be used for coarsening)
    for n, tempres in enumerate(residues):
        if tempres.get_has_structure():
            raise Exception("You tried to build ideal_helix for a residue "
                            "that already has structure:",tempres)
        if n>0 and (not tempres.get_index()==prev_idx+1):
            raise Exception("Passed non-contiguous segment to build_ideal_helix for",tempres.get_molecule())

        # New residue particle will replace the TempResidue's existing (empty) hierarchy
        rp = IMP.Particle(model)
        rp.set_name("Residue_%i" % tempres.get_index())

        # Copy the original residue type and index
        this_res = IMP.atom.Residue.setup_particle(rp,tempres.get_hierarchy())

        # Create the CAlpha
        ap = IMP.Particle(model)
        d = IMP.core.XYZR.setup_particle(ap)
        x = 2.3 * cos(n * 2 * pi / 3.6)
        y = 2.3 * sin(n * 2 * pi / 3.6)
        z = 6.2 / 3.6 / 2 * n * 2 * pi / 3.6
        d.set_coordinates(IMP.algebra.Vector3D(x, y, z))
        d.set_radius(1.0)
        a = IMP.atom.Atom.setup_particle(ap, IMP.atom.AT_CA)  # Decorating as Atom also decorates as Mass
        IMP.atom.Mass(ap).set_mass(110.0)
        this_res.add_child(a)

        # Add this structure to the TempResidue
        tempres.set_structure(this_res)
        created_hiers.append(this_res)
        prev_idx = tempres.get_index()
    coord_finder.add_residues(created_hiers) #the coord finder is for placing beads (later)


class Sequences(object):
    """A dictionary-like wrapper for reading and storing sequence data"""
    def __init__(self,fasta_fn,name_map=None):
        """read a fasta file and extract all the requested sequences
        @param fasta_fn sequence file
        @param name_map dictionary mapping the fasta name to final stored name
        """
        self.sequences = IMP.pmi1.tools.OrderedDict()
        self.read_sequences(fasta_fn,name_map)
    def __len__(self):
        return len(self.sequences)
    def __contains__(self,x):
        return x in self.sequences
    def __getitem__(self,key):
        if type(key) is int:
            try:
                allseqs = list(self.sequences.keys())
                return self.sequences[allseqs[key]]
            except:
                raise Exception("You tried to access sequence num",key,"but there's only",len(self.sequences.keys()))
        else:
            return self.sequences[key]
    def __iter__(self):
        return self.sequences.__iter__()
    def __repr__(self):
        ret=''
        for s in self.sequences:
            ret += '%s\t%s\n'%(s,self.sequences[s])
        return ret
    def read_sequences(self,fasta_fn,name_map=None):
        code = None
        seq = None
        with open(fasta_fn,'r') as fh:
            for (num, line) in enumerate(fh):
                if line.startswith('>'):
                    if seq is not None:
                        self.sequences[code] = seq.strip('*')
                    code = line.rstrip()[1:]
                    if name_map is not None:
                        try:
                            code = name_map[code]
                        except:
                            pass
                    seq = ''
                else:
                    line = line.rstrip()
                    if line: # Skip blank lines
                        if seq is None:
                            raise Exception( \
    "Found FASTA sequence before first header at line %d: %s" % (num + 1, line))
                        seq += line
        if seq is not None:
            self.sequences[code] = seq.strip('*')


class TopologyReader(object):
    """Automatically setup Sytem and Degrees of Freedom with a formatted text file.
    The file is read in and each part of the topology is stored as a
    ComponentTopology object for input into IMP::pmi1::macros::BuildSystem.
    The topology file should be in a simple pipe-delimited format:
    @code{.txt}
|molecule_name|color|fasta_fn|fasta_id|pdb_fn|chain|residue_range|pdb_offset|bead_size|em_residues_per_gaussian|rigid_body|super_rigid_body|chain_of_super_rigid_bodies|flags|
|Rpb1   |blue   |1WCM.fasta|1WCM:A|1WCM.pdb|A|1,1140   |0|10|0|1|1,3|1||
|Rpb1   |blue   |1WCM.fasta|1WCM:A|1WCM.pdb|A|1141,1274|0|10|0|2|1,3|1||
|Rpb1   |blue   |1WCM.fasta|1WCM:A|1WCM.pdb|A|1275,END |0|10|0|3|1,3|1||
|Rpb2   |red    |1WCM.fasta|1WCM:B|1WCM.pdb|B|all      |0|10|0|4|2,3|2||
|Rpb2.1 |green  |1WCM.fasta|1WCM:B|1WCM.pdb|B|all      |0|10|0|4|2,3|2||

    @endcode

    These are the fields you can enter:
    - `component_name`: Name of the component (chain). Serves as the parent
      hierarchy for this structure.
    - `color`: The color used in the output RMF file. Uses chimera names or R,G,B values
    - `fasta_fn`: Name of FASTA file containing this component.
    - `fasta_id`: String found in FASTA sequence header line. The sequence read
      from the file is assumed to be a protein sequence. If it should instead
      be treated as RNA or DNA, add an ',RNA' or ',DNA' suffix. For example,
      a `fasta_id` of 'myseq,RNA' will read the sequence 'myseq' from the
      FASTA file and treat it as RNA.
    - `pdb_fn`: Name of PDB file with coordinates (if available).
       If left empty, will set up as BEADS (you can also specify "BEADS")
       Can also write "IDEAL_HELIX".
    - `chain`: Chain ID of this domain in the PDB file.
    - `residue_range`: Comma delimited pair defining range.
       Can leave empty or use 'all' for entire sequence from PDB file.
       The second item in the pair can be END to select the last residue in the
       PDB chain.
    - `pdb_offset`: Offset to sync PDB residue numbering with FASTA numbering.
    - `bead_size`: The size (in residues) of beads used to model areas not
      covered by PDB coordinates. These will be automatically built.
    - `em_residues`: The number of Gaussians used to model the electron
      density of this domain. Set to zero if no EM fitting will be done.
      The GMM files will be written to <gmm_dir>/<component_name>_<em_res>.txt (and .mrc)
    - `rigid_body`: Leave empty if this object is not in a rigid body.
       Otherwise, this is a number corresponding to the rigid body containing
       this object. The number itself is just used for grouping things.
    - `super_rigid_body`: Like a rigid_body, except things are only occasionally rigid
    - `chain_of_super_rigid_bodies` For a polymer, create SRBs from groups.
    - `flags` additional flags for advanced options
    \note All filenames are relative to the paths specified in the constructor.

    """
    def __init__(self,
                 topology_file,
                 pdb_dir='./',
                 fasta_dir='./',
                 gmm_dir='./'):
        """Constructor.
        @param topology_file Pipe-delimited file specifying the topology
        @param pdb_dir Relative path to the pdb directory
        @param fasta_dir Relative path to the fasta directory
        @param gmm_dir Relative path to the GMM directory
        """
        self.topology_file = topology_file
        self.molecules = IMP.pmi1.tools.OrderedDict() # key=molname, value=TempMolecule
        self.pdb_dir = pdb_dir
        self.fasta_dir = fasta_dir
        self.gmm_dir = gmm_dir
        self._components = self.read(topology_file)

    # Preserve old self.component_list for backwards compatibility
    @IMP.deprecated_method("2.7",
                       "Use 'get_components()' instead of 'component_list'.")
    def __get_component_list(self): return self._components
    component_list = property(__get_component_list)

    def write_topology_file(self,outfile):
        with open(outfile, "w") as f:
            f.write("|molecule_name|color|fasta_fn|fasta_id|pdb_fn|chain|"
                    "residue_range|pdb_offset|bead_size|em_residues_per_gaussian|"
                    "rigid_body|super_rigid_body|chain_of_super_rigid_bodies|\n")
            for c in self._components:
                output = c.get_str()+'\n'
                f.write(output)
        return outfile

    def get_components(self, topology_list = "all"):
        """ Return list of ComponentTopologies for selected components
        @param topology_list List of indices to return"""
        if topology_list == "all":
            topologies = self._components
        else:
            topologies=[]
            for i in topology_list:
                topologies.append(self._components[i])
        return topologies

    def get_molecules(self):
        return self.molecules

    def read(self, topology_file, append=False):
        """Read system components from topology file. append=False will erase
        current topology and overwrite with new
        """
        is_topology = False
        is_directories = False
        linenum = 1
        if append==False:
            self._components=[]

        with open(topology_file) as infile:
            for line in infile:
                if line.lstrip()=="" or line[0]=="#":
                    continue
                elif line.split('|')[1].strip() in ("molecule_name"):
                    is_topology=True
                    is_directories = False
                    old_format = False
                    continue
                elif line.split('|')[1] == "component_name":
                    is_topology = True
                    IMP.handle_use_deprecated(
                          "Old-style topology format (using "
                          "|component_name|) is deprecated. Please switch to "
                          "the new-style format (using |molecule_name|)\n")
                    old_format = True
                    is_directories = False
                    continue
                elif line.split('|')[1] == "directories":
                    IMP.handle_use_deprecated(
                          "Setting directories in the topology file "
                          "is deprecated. Please do so through the "
                          "TopologyReader constructor. Note that new-style "
                          "paths are relative to the current working "
                          "directory, not the topology file.\n")
                    is_directories = True
                elif is_directories:
                    fields = line.split('|')
                    setattr(self, fields[1],
                            IMP.get_relative_path(topology_file, fields[2]))
                if is_topology:
                    new_component = self._parse_line(line, linenum, old_format)
                    self._components.append(new_component)
                    linenum += 1
        return self._components

    def _parse_line(self, component_line, linenum, old_format):
        """Parse a line of topology values and matches them to their key.
        Checks each value for correct syntax
        Returns a list of Component objects
        fields:
        """
        c = _Component()
        values = [s.strip() for s in component_line.split('|')]
        errors = []

        ### Required fields
        if old_format:
            c.molname = values[1]
            c.copyname = ''
            c._domain_name = values[2]
            c.color = 'blue'
        else:
            names = values[1].split('.')
            if len(names)==1:
                c.molname = names[0]
                c.copyname = ''
            elif len(names)==2:
                c.molname = names[0]
                c.copyname = names[1]
            else:
                c.molname = names[0]
                c.copyname = names[1]
                errors.append("Molecule name should be <molecule.copyID>")
                errors.append("For component %s line %d " % (c.molname,linenum))
            c._domain_name = c.molname + '.' + c.copyname
            colorfields = values[2].split(',')
            if len(colorfields)==3:
                c.color = [float(x) for x in colorfields]
                if any([x>1 for x in c.color]):
                    c.color=[x/255 for x in c.color]
            else:
                c.color = values[2]
        c._orig_fasta_file = values[3]
        c.fasta_file = values[3]
        fasta_field = values[4].split(",")
        c.fasta_id  = fasta_field[0]
        c.fasta_flag = None
        if len(fasta_field) > 1:
            c.fasta_flag = fasta_field[1]
        c._orig_pdb_input = values[5]
        pdb_input = values[5]
        tmp_chain = values[6]
        rr = values[7]
        offset = values[8]
        bead_size = values[9]
        emg = values[10]
        if old_format:
            rbs = srbs = csrbs = ''
        else:
            rbs = values[11]
            srbs = values[12]
            csrbs = values[13]

        if c.molname not in self.molecules:
            self.molecules[c.molname] = _TempMolecule(c)
        else:
            # COPY OR DOMAIN
            c._orig_fasta_file = self.molecules[c.molname].orig_component.fasta_file
            c.fasta_id = self.molecules[c.molname].orig_component.fasta_id
            self.molecules[c.molname].add_component(c,c.copyname)

        # now cleanup input
        c.fasta_file = os.path.join(self.fasta_dir,c._orig_fasta_file)
        if pdb_input=="":
            errors.append("PDB must have BEADS, IDEAL_HELIX, or filename")
            errors.append("For component %s line %d is not correct"
                          "|%s| was given." % (c.molname,linenum,pdb_input))
        elif pdb_input in ("IDEAL_HELIX","BEADS"):
            c.pdb_file = pdb_input
        else:
            c.pdb_file = os.path.join(self.pdb_dir,pdb_input)

            # PDB chain must be one or two characters
            if len(tmp_chain)==1 or len(tmp_chain)==2:
                c.chain = tmp_chain
            else:
                errors.append("PDB Chain identifier must be one or two characters.")
                errors.append("For component %s line %d is not correct"
                              "|%s| was given." % (c.molname,linenum,tmp_chain))

        ### Optional fields
        # Residue Range
        if rr.strip()=='all' or str(rr)=="":
            c.residue_range = None
        elif len(rr.split(','))==2 and self._is_int(rr.split(',')[0]) and (self._is_int(rr.split(',')[1]) or rr.split(',')[1] == 'END'):
            # Make sure that is residue range is given, there are only two values and they are integers
            c.residue_range = (int(rr.split(',')[0]), rr.split(',')[1])
            if c.residue_range[1] != 'END':
                c.residue_range = (c.residue_range[0], int(c.residue_range[1]))
            # Old format used -1 for the last residue
            if old_format and c.residue_range[1] == -1:
                c.residue_range = (c.residue_range[0], 'END')
        else:
            errors.append("Residue Range format for component %s line %d is not correct" % (c.molname, linenum))
            errors.append("Correct syntax is two comma separated integers:  |start_res, end_res|. end_res can also be END to select the last residue in the chain. |%s| was given." % rr)
            errors.append("To select all residues, indicate |\"all\"|")

        # PDB Offset
        if self._is_int(offset):
            c.pdb_offset=int(offset)
        elif len(offset)==0:
            c.pdb_offset = 0
        else:
            errors.append("PDB Offset format for component %s line %d is not correct" % (c.molname, linenum))
            errors.append("The value must be a single integer. |%s| was given." % offset)

        # Bead Size
        if self._is_int(bead_size):
            c.bead_size=int(bead_size)
        elif len(bead_size)==0:
            c.bead_size = 0
        else:
            errors.append("Bead Size format for component %s line %d is not correct" % (c.molname, linenum))
            errors.append("The value must be a single integer. |%s| was given." % bead_size)

        # EM Residues Per Gaussian
        if self._is_int(emg):
            if int(emg) > 0:
                c.density_prefix = os.path.join(self.gmm_dir,c.get_unique_name())
                c.gmm_file = c.density_prefix+'.txt'
                c.mrc_file = c.density_prefix+'.gmm'

                c.em_residues_per_gaussian=int(emg)
            else:
                c.em_residues_per_gaussian = 0
        elif len(emg)==0:
            c.em_residues_per_gaussian = 0
        else:
            errors.append("em_residues_per_gaussian format for component "
                          "%s line %d is not correct" % (c.molname, linenum))
            errors.append("The value must be a single integer. |%s| was given." % emg)

        # rigid bodies
        if len(rbs)>0:
            if not self._is_int(rbs):
                errors.append("rigid bodies format for component "
                              "%s line %d is not correct" % (c.molname, linenum))
                errors.append("Each RB must be a single integer, or empty. "
                              "|%s| was given." % rbs)
            c.rigid_body = int(rbs)

        # super rigid bodies
        if len(srbs)>0:
            srbs = srbs.split(',')
            for i in srbs:
                if not self._is_int(i):
                    errors.append("super rigid bodies format for component "
                                  "%s line %d is not correct" % (c.molname, linenum))
                    errors.append("Each SRB must be a single integer. |%s| was given." % srbs)
            c.super_rigid_bodies = srbs

        # chain of super rigid bodies
        if len(csrbs)>0:
            if not self._is_int(csrbs):
                errors.append("em_residues_per_gaussian format for component "
                              "%s line %d is not correct" % (c.molname, linenum))
                errors.append("Each CSRB must be a single integer. |%s| was given." % csrbs)
            c.chain_of_super_rigid_bodies = csrbs

        # done
        if errors:
            raise ValueError("Fix Topology File syntax errors and rerun: " \
                             + "\n".join(errors))
        else:
            return c


    def set_gmm_dir(self,gmm_dir):
        """Change the GMM dir"""
        self.gmm_dir = gmm_dir
        for c in self._components:
            c.gmm_file = os.path.join(self.gmm_dir,c.get_unique_name()+".txt")
            c.mrc_file = os.path.join(self.gmm_dir,c.get_unique_name()+".mrc")
            print('new gmm',c.gmm_file)

    def set_pdb_dir(self,pdb_dir):
        """Change the PDB dir"""
        self.pdb_dir = pdb_dir
        for c in self._components:
            if not c._orig_pdb_input in ("","None","IDEAL_HELIX","BEADS"):
                c.pdb_file = os.path.join(self.pdb_dir,c._orig_pdb_input)

    def set_fasta_dir(self,fasta_dir):
        """Change the FASTA dir"""
        self.fasta_dir = fasta_dir
        for c in self._components:
            c.fasta_file = os.path.join(self.fasta_dir,c._orig_fasta_file)

    def _is_int(self, s):
       # is this string an integer?
        try:
            float(s)
            return float(s).is_integer()
        except ValueError:
            return False

    def get_rigid_bodies(self):
        """Return list of lists of rigid bodies (as domain name)"""
        rbl = defaultdict(list)
        for c in self._components:
            if c.rigid_body:
                rbl[c.rigid_body].append(c.get_unique_name())
        return rbl.values()

    def get_super_rigid_bodies(self):
        """Return list of lists of super rigid bodies (as domain name)"""
        rbl = defaultdict(list)
        for c in self._components:
            for rbnum in c.super_rigid_bodies:
                rbl[rbnum].append(c.get_unique_name())
        return rbl.values()

    def get_chains_of_super_rigid_bodies(self):
        """Return list of lists of chains of super rigid bodies (as domain name)"""
        rbl = defaultdict(list)
        for c in self._components:
            for rbnum in c.chain_of_super_rigid_bodies:
                rbl[rbnum].append(c.get_unique_name())
        return rbl.values()

class _TempMolecule(object):
    """Store the Components and any requests for copies"""
    def __init__(self,init_c):
        self.molname = init_c.molname
         # key=copy ID, value = list of domains
        self.domains = IMP.pmi1.tools.OrderedDefaultDict(list)
        self.add_component(init_c,init_c.copyname)
        self.orig_copyname = init_c.copyname
        self.orig_component = self.domains[init_c.copyname][0]
    def add_component(self,component,copy_id):
        self.domains[copy_id].append(component)
        component.domainnum = len(self.domains[copy_id])-1
    def __repr__(self):
        return ','.join('%s:%i'%(k,len(self.domains[k])) for k in self.domains)

class _Component(object):
    """Stores the components required to build a standard IMP hierarchy
    using IMP.pmi1.BuildModel()
    """
    def __init__(self):
        self.molname = None
        self.copyname = None
        self.domainnum = 0
        self.fasta_file = None
        self._orig_fasta_file = None
        self.fasta_id = None
        self.fasta_flag = None
        self.pdb_file = None
        self._orig_pdb_input = None
        self.chain = None
        self.residue_range = None
        self.pdb_offset = 0
        self.bead_size = 10
        self.em_residues_per_gaussian = 0
        self.gmm_file = ''
        self.mrc_file = ''
        self.density_prefix = ''
        self.color = 0.1
        self.rigid_body = None
        self.super_rigid_bodies = []
        self.chain_of_super_rigid_bodies = []

    def _l2s(self,l):
        return ",".join("%s" % x for x in l)

    def __repr__(self):
        return self.get_str()

    def get_unique_name(self):
        return "%s.%s.%i"%(self.molname,self.copyname,self.domainnum)

    def get_str(self):
        res_range = self.residue_range
        if self.residue_range is None:
            res_range = []
        name = self.molname
        if self.copyname!='':
            name += '.'+self.copyname
        if self.chain is None:
            chain = ' '
        else:
            chain = self.chain
        color=self.color
        if isinstance(color, list):
            color=','.join([str(x) for x in color])
        a= '|'+'|'.join([name,color,self._orig_fasta_file,self.fasta_id,
                         self._orig_pdb_input,chain,self._l2s(list(res_range)),
                             str(self.pdb_offset),str(self.bead_size),
                             str(self.em_residues_per_gaussian),
                             str(self.rigid_body) if self.rigid_body else '',
                             self._l2s(self.super_rigid_bodies),
                             self._l2s(self.chain_of_super_rigid_bodies)])+'|'
        return a

    # Preserve old self.name for backwards compatibility
    @IMP.deprecated_method("2.7", "Use 'molname' instead of 'name'.")
    def __get_name(self): return self.molname
    name = property(__get_name)

    # Preserve old self.domain_name for backwards compatibility
    @IMP.deprecated_method("2.7",
                           "Use 'get_unique_name()' instead of 'domain_name'.")
    def __get_domain_name(self): return self._domain_name
    domain_name = property(__get_domain_name)


class PMIMoleculeHierarchy(IMP.atom.Molecule):
    '''Extends the functionality of IMP.atom.Molecule'''

    def __init__(self,hierarchy):
        IMP.atom.Molecule.__init__(self,hierarchy)

    def get_state_index(self):
        state = self.get_parent()
        return IMP.atom.State(state).get_state_index()

    def get_copy_index(self):
        return IMP.atom.Copy(self).get_copy_index()

    def get_extended_name(self):
        return self.get_name()+"."+\
               str(self.get_copy_index())+\
               "."+str(self.get_state_index())

    def get_sequence(self):
        return IMP.atom.Chain(self).get_sequence()

    def get_residue_indexes(self):
        return IMP.pmi1.tools.get_residue_indexes(self)

    def get_residue_segments(self):
        return IMP.pmi1.tools.Segments(self.get_residue_indexes())

    def get_chain_id(self):
        return IMP.atom.Chain(self).get_id()

    def __repr__(self):
        s='PMIMoleculeHierarchy '
        s+=self.get_name()
        s+=" "+"Copy  "+str(IMP.atom.Copy(self).get_copy_index())
        s+=" "+"State "+str(self.get_state_index())
        s+=" "+"N residues "+str(len(self.get_sequence()))
        return s



#!/usr/bin/env python

"""Add navigation to the reference guide HTML output, similarly
   to what is done for the manual (see doxygen_add_nav.py).
"""

from __future__ import print_function
from doxygen_add_nav import Docs, Page
import tools
import sys
import os


def all_module_docs(modules):
    return [os.path.join('modules', m, 'doc') for m in modules]


def get_module_readme(module):
    """Make a Page object for a module's top-level README file"""
    modulesub = module.replace('_', '__')
    p = Page('namespace' + modulesub)
    if module == 'kernel':
        p.out_file_name = 'namespaceIMP'
    else:
        p.out_file_name = 'namespaceIMP_1_1' + modulesub
    p.source_file_name = 'modules/%s/README.md' % module
    return p


def main():
    source = sys.argv[1]
    modules = [m[0] for m in tools.get_modules(source)]
    docs = Docs(xml_dir='doxygen/ref/xml',
                html_dir='doc/ref',
                top_source_dir=source,
                source_subdirs=sys.argv[2:] + all_module_docs(modules))
    for p in docs.pages:
        p.map_location_to_source(source)
    for m in modules:
        docs.pages.append(get_module_readme(m))

    # Difficult to figure out the main/index page, so set this manually
    p = docs.page_by_id['indexpage']
    p.out_file_name = 'index'
    p.source_file_name = 'doc/ref/mainpage.md'

    for p in docs.pages:
        if docs.get_html_pages(p):
            docs.add_page_navigation(p)
        else:
            print("Not handling %s: HTML file does not exist" % p.id)


if __name__ == '__main__':
    main()
